
# Built-in
import os
import warnings
import functools as ftools

# Common
import numpy as np

# tofu
pfe = os.path.join(os.path.expanduser('~'), '.tofu', '_imas2tofu_def.py')
if os.path.isfile(pfe):
    # Make sure we load the user-specific file
    # sys.path method
    # sys.path.insert(1, os.path.join(os.path.expanduser('~'), '.tofu'))
    # import _scripts_def as _defscripts
    # _ = sys.path.pop(1)
    # importlib method
    import importlib.util
    spec = importlib.util.spec_from_file_location("_defimas2tofu", pfe)
    _defimas2tofu = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(_defimas2tofu)
else:
    try:
        import tofu.imas2tofu._def as _defimas2tofu
    except Exception as err:
        from . import _def as _defimas2tofu

# imas
try:
    import imas
except Exception as err:
    raise Exception('imas not available')


# Useful scalar types
_NINT = (np.int32, np.int64)
_INT = (int,) + _NINT
_NFLOAT = (np.float32, np.float64)
_FLOAT = (float,) + _NFLOAT
_NUMB = _INT + _FLOAT
_BOOL = (bool, np.bool_)


_DSHORT = _defimas2tofu._dshort
_DCOMP = _defimas2tofu._dcomp
_DDUNITS = imas.dd_units.DataDictionaryUnits()

_ISCLOSE = True
_POS = False
_EMPTY = True
_NAN = True
_DATA = True
_UNITS = True
_STRICT = True
_STACK = True
_WARN = True
_RETURN_ALL = False


# #############################################################################
#                      Units functions
# #############################################################################


def _prepare_sig_units(sig, units=False):
    """ Remove brackest from shortcut, if any """
    if '[' in sig:
        while '[' in sig:
            sig = sig[:sig.index('[')] + sig[sig.index(']')+1:]
    return sig


def get_units(ids, sig, dshort=None, dcomp=None, force=None):
    """ Get units from imas.dd_units.DataDictionaryUnits() """
    if dshort is None:
        dshort = _DSHORT
    if dcomp is None:
        dcomp = _DCOMP
    if force is None:
        force = True
    if sig in dshort[ids].keys():
        sig = _prepare_sig_units(dshort[ids][sig]['str'])
    else:
        sig = _prepare_sig_units(sig)
    units = _DDUNITS.get_units(ids, sig.replace('.', '/'))

    # Condition in which to use tofu units instead of imas units
    c0 = (units is None
          and force is True
          and (sig in dshort[ids].keys() or sig in dcomp[ids].keys()))
    if c0 is True:
        if sig in dshort[ids].keys():
            tofuunits = dshort[ids][sig].get('units')
        else:
            tofuunits = dcomp[ids][sig].get('units')
        if tofuunits != units:
            units = tofuunits
    return units


# #############################################################################
#                      Data retrieveing function
# #############################################################################


def _prepare_sig(sig):
    if '[' in sig:
        # Get nb and ind
        ind0 = 0
        while '[' in sig[ind0:]:
            ind1 = ind0 + sig[ind0:].index('[')
            ind2 = ind0 + sig[ind0:].index(']')
            sig = sig.replace(sig[ind1+1:ind2],
                              sig[ind1+1:ind2].replace('.', '/'))
            ind0 = ind2 + 1
    return sig


def _get_condfromstr(sid, sig=None):
    lid0, id1 = sid.split('=')
    lid0 = lid0.split('.')

    if '.' in id1 and id1.replace('.', '').isdecimal():
        id1 = float(id1)
    elif id1.isdecimal():
        id1 = int(id1)
    elif '.' in id1:
        msg = ("Not clear how to interpret the following condition:\n"
               + "\t- sig: {}\n".format(sig)
               + "\t- condition: {}".format(sid))
        raise Exception(msg)
    return lid0, id1


def get_fsig(sig):
    # break sig in list of elementary nodes
    sig = _prepare_sig(sig)
    ls0 = sig.split('.')
    sig = sig.replace('/', '.')
    ls0 = [ss.replace('/', '.') for ss in ls0]
    ns = len(ls0)

    # For each node, identify type (i.e. [])
    lc = [all([si in ss for si in ['[', ']']]) for ss in ls0]
    dcond, seq, nseq, jj = {}, [], 0, 0
    for ii in range(0, ns):
        nseq = len(seq)
        if lc[ii]:
            # there is []
            if nseq > 0:
                dcond[jj] = {'type': 0, 'lstr': seq}
                seq = []
                jj += 1

            # Isolate [strin]
            ss = ls0[ii]
            strin = ss[ss.index('[')+1:-1]

            # typ 0 => no dependency
            # typ 1 => dependency ([],[time],[chan],[int])
            # typ 2 => selection ([...=...])
            cond, ind, typ = None, None, 1
            if '=' in strin:
                typ = 2
                cond = _get_condfromstr(strin, sig=sig)
            elif strin in ['time', 'chan']:
                ind = strin
            elif strin.isnumeric():
                ind = [int(strin)]
            dcond[jj] = {'str': ss[:ss.index('[')], 'type': typ,
                         'ind': ind, 'cond': cond}
            jj += 1
        else:
            seq.append(ls0[ii])
            if ii == ns-1:
                dcond[jj] = {'type': 0, 'lstr': seq}

    c0 = [v['type'] == 1 and (v['ind'] is None or len(v['ind']) > 1)
          for v in dcond.values()]
    if np.sum(c0) > 1:
        msg = ("Cannot handle mutiple iterative levels yet !\n"
               + "\t- sig: {}".format(sig))
        raise Exception(msg)

    # Create function for getting signal
    def fsig(obj, indt=None, indch=None, stack=None, dcond=dcond):
        if stack is None:
            stack = _STACK
        sig = [obj]
        nsig = 1
        for ii in dcond.keys():

            # Standard case (no [])
            if dcond[ii]['type'] == 0:
                sig = [ftools.reduce(getattr, [sig[jj]]+dcond[ii]['lstr'])
                       for jj in range(0, nsig)]

            # dependency
            elif dcond[ii]['type'] == 1:
                for jj in range(0, nsig):
                    sig[jj] = getattr(sig[jj], dcond[ii]['str'])
                    nb = len(sig[jj])
                    if dcond[ii]['ind'] == 'time':
                        ind = indt
                    elif dcond[ii]['ind'] == 'chan':
                        ind = indch
                    else:
                        ind = dcond[ii]['ind']

                    if ind is None:
                        ind = range(0, nb)

                    if nsig > 1:
                        if isinstance(ind, str) or len(ind) != 1:
                            msg = ('ind should be have len() = 1\n'
                                   + '\t- ind: {}'.format(ind))
                            raise Exception(msg)

                    if len(ind) == 1:
                        if len(sig[jj]) < ind[0] + 1:
                            msg = (
                                f"dcond[{ii}]['ind'] = {dcond[ii]['ind']} "
                                f"so ind = {ind} "
                                f"but len(sig[{jj}]) = {len(sig[jj])}"
                            )
                            raise Exception(msg)
                        else:
                            sig[jj] = sig[jj][ind[0]]
                    else:
                        if nsig != 1:
                            msg = ("nsig should be 1!\n"
                                   + "\t- nsig: {}".format(nsig))
                            raise Exception(msg)
                        sig = [sig[0][ll] for ll in ind]
                        nsig = len(sig)

            # one index to be found
            else:
                for jj in range(0, nsig):
                    sig[jj] = getattr(sig[jj], dcond[ii]['str'])
                    nb = len(sig[jj])
                    typ = type(ftools.reduce(
                        getattr, [sig[jj][0]] + dcond[ii]['cond'][0]))
                    if typ == str:
                        ind = [
                            ll for ll in range(0, nb)
                            if (ftools.reduce(
                                getattr,
                                [sig[jj][ll]] + dcond[ii]['cond'][0]).strip()
                                == dcond[ii]['cond'][1].strip())]
                    else:
                        ind = [ll for ll in range(0, nb)
                               if (ftools.reduce(
                                   getattr,
                                   [sig[jj][ll]]+dcond[ii]['cond'][0])
                                   == dcond[ii]['cond'][1])]
                    if len(ind) != 1:
                        msg = ("No / several matching signals for:\n"
                               + "\t- {}[]{} = {}\n".format(
                                   dcond[ii]['str'],
                                   dcond[ii]['cond'][0],
                                   dcond[ii]['cond'][1])
                               + "\t- nb.of matches: {}".format(len(ind)))
                        raise Exception(msg)
                    sig[jj] = sig[jj][ind[0]]

        # Conditions for stacking / sqeezing sig
        lc = [
            (
                stack and nsig > 1 and isinstance(sig[0], np.ndarray)
                and all([ss.shape == sig[0].shape for ss in sig[1:]])
            ),
            (
                stack and nsig > 1
                and isinstance(sig[0], _NUMB + (str,))
            ),
            (
                stack and nsig == 1
                and type(sig) in [np.ndarray, list, tuple]
            ),
        ]

        if lc[0]:
            sig = np.atleast_1d(np.squeeze(np.stack(sig)))
        elif lc[1] or lc[2]:
            sig = np.atleast_1d(np.squeeze(sig))
        return sig
    return fsig


# #############################################################################
#                      Data functions
# #############################################################################


def _checkformat_getdata_dsig(
    dsig=None,
    occ=None,
    indch=None,
    indt=None,
    dids=None,
    dshort=None,
    dcomp=None,
    dall_except=None
):
    """ Check the desired ids / signal is available """

    if not isinstance(dids, dict):
        msg = ("dids must be a dict\n"
               + "\t- provided: {}".format(type(dids)))
        raise Exception(msg)

    lc = [dsig is None,
          isinstance(dsig, str) and dsig in dids.keys(),
          isinstance(dsig, list) and all([ss in dids.keys() for ss in dsig]),
          isinstance(dsig, dict)
          and all([k0 in dids.keys()
                   and (v0 is None
                        or isinstance(v0, str)
                        or (isinstance(v0, list)
                            and all([isinstance(ss, str) for ss in v0]))
                        or isinstance(v0, dict))
                   for k0, v0 in dsig.items()])]
    if not any(lc):
        msg = ("Arg dsig must be either:\n"
               + "\t- None: All shortucts of all ids\n"
               + "\t- str: a valid ids, all shortcuts\n"
               + "\t- list: a list of valid ids, all shortcuts for each\n"
               + "\t- dict: a dict of the form:\n"
               + "\t        {ids0: [short00, ..., short0N],\n"
               + "\t         ids1: [short10, ..., short1M],\n"
               + "\t         ids2: short2}\n"
               + "\t    Where the values are shortucts or list of such\n\n"
               + "  Provided : {}\n".format(dids)
               + "  Available: {}\n".format(list(dids.keys()))
               + "  => Consider using self.add_ids({})".format(str(dids)))
        raise Exception(msg)

    if lc[0]:
        dsig = dict.fromkeys(dids.keys())
    elif lc[1]:
        dsig = dict.fromkeys([dsig])
    elif lc[2]:
        dsig = dict.fromkeys(dsig)
    elif lc[3]:
        # copy to avoid reference
        dsig = dict(dsig)

    # Check occurences, channels and signals
    for k0 in dsig.keys():

        # Extract occ, indch and sig if any
        lc = [dsig[k0] is None,
              isinstance(dsig[k0], str),
              isinstance(dsig[k0], list),
              isinstance(dsig[k0], dict)]
        if not any(lc):
            msg = ""
            raise Exception(msg)

        if lc[0] or lc[1] or lc[2]:
            dsig[k0] = {'sig': dsig[k0]}
            occi, indchi, indti, sig = occ, indch, indt, dsig[k0]['sig']
        elif lc[3]:
            occi = dsig['k0'].get('occ', occ)
            indchi = dsig[k0].get('indch', indch)
            indti = dsig[k0].get('indt', indt)
            sig = dsig[k0].get('sig')

        # Check occ
        occi = _checkformat_getdata_occ(occi, k0, dids=dids)
        indoc = np.where(dids[k0]['occ'] == occi)[0]

        # Check all occ have isget = True
        indok = dids[k0]['isget'][indoc]
        if not np.all(indok):
            msg = ("All desired occurences shall have not been gotten!\n"
                   + "    - desired occ:   {}\n".format(occi)
                   + "    - available occ: {}\n".format(dids[k0]['occ'])
                   + "    - isget:         {}".format(dids[k0]['isget'])
                   + "\n  => Try running self.open_get_close()")
            raise Exception(msg)

        # Check indch
        if hasattr(dids[k0]['ids'][indoc[0]], 'channel'):
            nch = len(getattr(dids[k0]['ids'][indoc[0]], 'channel'))
            indchi = _checkformat_getdata_indch(indchi, nch)

        # Check shortcuts
        sig, comp = _checkformat_getdata_sig(sig, k0,
                                             dshort=dshort, dcomp=dcomp,
                                             dall_except=dall_except)
        dsig[k0] = {'occ': occi, 'indch': indchi, 'indt': indti,
                    'sig': sig, 'comp': comp}
    return dsig


def _checkformat_getdata_sig(sig, ids,
                             dshort=None, dcomp=None, dall_except=None):
    lc = [isinstance(dd, dict) for dd in [dshort, dcomp, dall_except]]
    if not all(lc):
        msg = ("dshort, dcomp and dall_except must be dict:\n"
               + "\t- type(dshort): {}\n".format(type(dshort))
               + "\t- type(dcomp): {}\n".format(type(dcomp))
               + "\t- type(dall_except): {}\n".format(type(dall_except)))
        raise Exception(msg)

    msg = ("Arg sig must be a str or a list of str!\n"
           + "  More specifically, a list of valid ids nodes paths")
    lks = list(dshort[ids].keys())
    lkc = list(dcomp[ids].keys())
    lk = set(lks).union(lkc)

    if ids in dall_except.keys():
        lk = lk.difference(dall_except[ids])

    lc = [sig is None, type(sig) is str, type(sig) is list]
    if not any(lc):
        raise Exception(msg)
    if lc[0]:
        sig = list(lk)
    elif lc[1]:
        sig = [sig]
    elif lc[2]:
        if any([type(ss) is not str for ss in sig]):
            raise Exception(msg)
    nsig = len(sig)

    # Check each sig is either a key / value[str] to dshort
    comp = np.zeros((nsig,), dtype=bool)
    for ii in range(0, nsig):
        lc0 = [sig[ii] in lks,
               [sig[ii] == dshort[ids][kk]['str'] for kk in lks]]
        c1 = sig[ii] in lkc
        if not (lc0[0] or any(lc0[1]) or c1):
            msg = ("Each provided sig must be either:\n"
                   + "\t- a valid shortcut (cf. self.shortcuts()\n"
                   + "\t- a valid long version (cf. self.shortcuts)\n"
                   + "\n  Provided sig: {} for ids {}".format(sig[ii], ids))
            raise Exception(msg)
        if c1:
            comp[ii] = True
        else:
            if not lc0[0]:
                sig[ii] = lks[lc0[1].index(True)]
    return sig, comp


def _checkformat_getdata_occ(occ, ids, dids=None):
    if not isinstance(dids, dict):
        msg = ("dids must be a dict\n"
               + "\t- provided: {}".format(type(dids)))
        raise Exception(msg)

    msg = ("Arg occ must be a either:\n"
           + "\t- None: all occurences are used\n"
           + "\t- int: occurence (in self.dids[{}]['occ'])\n".format(ids)
           + "\t- int array: occurences (in self.dids[{}]['occ'])".format(ids))
    lc = [occ is None,
          type(occ) is int,
          hasattr(occ, '__iter__')]

    if not any(lc):
        raise Exception(msg)
    if lc[0]:
        occ = dids[ids]['occ']
    else:
        occ = np.r_[occ].astype(int).ravel()
        if any([oc not in dids[ids]['occ'] for oc in occ]):
            raise Exception(msg)
    return occ


def _checkformat_getdata_indch(indch, nch):
    """ Check index of channels, returns array of int indices

    Parameters
    ----------
    indch : None / int / aiterable of int or bool
        Input index of channels
    nch : int
        Max number of channels

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    indch : np.ndarray of int
        Output indices

    """

    # -----------------------------
    # Initialize error msg
    # -----------------------------

    msg = (
        "Arg indch must be a either:\n"
        "\t- None: all channels used\n"
        "\t- int: channel to use (index)\n"
        "\t- array of int: channels to use (indices)\n"
        "\t- array of bool: channels to use (indices)\n"
    )

    # -----------------------------
    # List of acceptable conditions
    # -----------------------------

    lc0 = [
        indch is None,
        isinstance(indch, _INT),
        hasattr(indch, '__iter__') and not isinstance(indch, str),
    ]

    if not any(lc0):
        raise Exception(msg)

    # ------------------------
    # None
    # ------------------------

    if lc0[0]:
        # defaul to all indices
        indch = np.arange(0, nch)

    # ------------------------
    # integer
    # ------------------------

    elif lc0[1]:
        # make numpy int array
        indch = np.r_[indch].ravel()

    # ------------------------
    # iterable: int or bool
    # ------------------------

    elif lc0[2]:
        # make numpy array
        indch = np.r_[indch].ravel()

        # get dtype (may also be a list)
        lc1 = [
            isinstance(indch[0], _INT),
            isinstance(indch[0], _BOOL),
        ]

        if not any(lc1):
            raise Exception(msg)

        if lc1[1]:
            # convert from bool to int
            indch = np.nonzero(indch)[0]

        # safety check
        if not np.all((indch >= 0) & (indch < nch)):
            msg = (
                "Some channel indices are out of scope!\n"
                f"\t- nch: {nch}\n"
                f"\t- indch: {indch}"
            )
            raise Exception(msg)

    return indch


def _check_data(data, pos=None, nan=None, isclose=None, empty=None):
    """ Check the data loaded from imas against several safety checks

    For each occurence

    Available checks:
        - only positive values
        - default IMAS values (>1e30)
        - duplicated unique vector
        - empty data (shape contains 0 oir size 0 or only nan)

    """

    # ------------
    # Check inputs
    if pos is None:
        pos = _POS
    if nan is None:
        nan = _NAN
    if isclose is None:
        isclose = _ISCLOSE
    if empty is None:
        empty = _EMPTY

    # ------------
    # Run checks on data

    # If isclose, check data contains a replicated vector (keep vector only)
    if isclose is True:
        for ii in range(0, len(data)):
            if isinstance(data[ii], np.ndarray) and data[ii].ndim == 2:
                if np.allclose(data[ii], data[ii][0:1, :]):
                    data[ii] = data[ii][0, :]
                elif np.allclose(data[ii], data[ii][:, 0:1]):
                    data[ii] = data[ii][:, 0]

    # All values larger than 1e30 are default imas values => nan
    if nan is True:
        for ii in range(0, len(data)):
            c0 = (
                isinstance(data[ii], np.ndarray)
                and data[ii].dtype in _NUMB
            )
            if c0 is True:
                # Make sure to test only non-nan to avoid warning
                ind = (~np.isnan(data[ii])).nonzero()
                ind2 = np.abs(data[ii][ind]) > 1.e30
                if np.any(ind2):
                    ind = tuple([ii[ind2] for ii in ind])
                    if data[ii].dtype in _INT:
                        data[ii] = data[ii].astype(float)
                    data[ii][ind] = np.nan

    # data supposed to be positive only (nan otherwise)
    if pos is True:
        for ii in range(0, len(data)):
            if isinstance(data[ii], np.ndarray):
                data[ii][data[ii] < 0] = np.nan

    # data appears to be empty or all nan
    isempty = [None for ii in range(len(data))]
    if empty is True:
        for ii in range(len(data)):
            isempty[ii] = (len(data[ii]) == 0
                           or (isinstance(data[ii], np.ndarray)
                               and (data[ii].size == 0
                                    or 0 in data[ii].shape)))
            if isinstance(data[ii], np.ndarray) and data[ii].dtype.kind != 'U':
                isempty[ii] &= bool(np.all(np.isnan(data[ii])))
    return data, isempty


def _get_data_units(ids=None, sig=None, occ=None,
                    comp=None, indt=None, indch=None,
                    stack=None, isclose=None, flatocc=None,
                    nan=None, pos=None, empty=None,
                    dids=None, dcomp=None, dshort=None, dall_except=None,
                    data=True, units=True, strict=None):
    """ Reference method for getting data and units, using shortcuts

    For a given ids, sig (shortcut) and occurence (occ)
    Get data if data = True
    Get units if units = True

    The data is:
        - restricted to the desired time indices (indt)
        - restricted to the desired channels (indch)
        - stacked into a single array if stack = True
        - checked whether many vectors are duplicates (expected)
            in which case a single vector is kept
        - checked for NaNs if nan = True
        - checked for positive values if pos = True
        - All occurences are flatened if flatocc = True

    If comp = True, it means the data is not loaded directly from the ids,
    but computed from a combination of signals in the ids.
    The computation function must have been defined previously

    """

    # --------------
    # Check inputs

    if data is None:
        data = _DATA
    if units is None:
        units = _UNITS

    # --------------
    # Load data

    # get list of results for occ
    occref = dids[ids]['occ']
    indoc = np.array([np.nonzero(occref == oc)[0][0] for oc in occ])
    nocc = len(indoc)
    out, unit, errdata, errunits = None, None, None, None

    # Data has te be computed
    if comp is True:
        if data is True:
            if stack is None:
                stack = True
            try:
                lstr = dcomp[ids][sig]['lstr']
                kargs = dcomp[ids][sig].get('kargs', {})
                ddata = get_data_units(
                    dsig={ids: lstr}, occ=occ, indch=indch,
                    data=True, units=False, indt=indt, stack=stack,
                    flatocc=False, nan=nan, pos=pos, warn=False,
                    dids=dids, dcomp=dcomp, dshort=dshort,
                    dall_except=dall_except,
                    strict=strict,
                )[ids]
                out = [dcomp[ids][sig]['func'](
                    *[ddata[kk]['data'][nn] for kk in lstr],
                    **kargs)
                       for nn in range(0, nocc)]
                if pos is None:
                    pos = dcomp[ids][sig].get('pos', False)
            except Exception as err:
                errdata = err

        if units is True:
            try:
                unit = dcomp[ids][sig].get('units', None)
            except Exception as err:
                errunits = err

    # Data available from ids
    else:
        if data is True:
            try:
                out = [dshort[ids][sig]['fsig'](dids[ids]['ids'][ii],
                                                indt=indt, indch=indch,
                                                stack=stack)
                       for ii in indoc]
                if pos is None:
                    pos = dshort[ids][sig].get('pos', False)
            except Exception as err:
                errdata = err

        if units is True:
            try:
                unit = get_units(ids, sig)
            except Exception as err:
                errunits = err

    # Check data
    isempty = None
    if errdata is None and data is True:
        out, isempty = _check_data(out,
                                   pos=pos, nan=nan,
                                   isclose=isclose, empty=empty)
        if np.all(isempty):
            msg = ("empty data in {}.{}".format(ids, sig))
            errdata = Exception(msg)
        elif nocc == 1 and flatocc is True:
            out = out[0]
            isempty = isempty[0]
    return {'data': out, 'units': unit,
            'isempty': isempty, 'errdata': errdata, 'errunits': errunits}


def get_data_units(dsig=None, occ=None,
                   data=None, units=None,
                   indch=None, indt=None, stack=None,
                   isclose=None, flatocc=True,
                   nan=True, pos=None, empty=None, strict=None,
                   return_all=None, warn=None,
                   dids=None, dshort=None, dcomp=None, dall_except=None):
    """ Return a dict with the data and units (and empty, errors)

    Can be used:
        - For multiple (all) shorcuts from a unique ids
        - For all shortcuts from multiple ids
        - For a custom ids: shortcuts dict

    For the desired ids and signal (shortcut), load the data and units
    Return a dict also containing a bool indicating whether the data is empty
    and error strings if any was raised

    """

    # ------------------
    # Check format input

    if data is None:
        data = _DATA
    if units is None:
        units = _UNITS
    if strict is None:
        strict = _STRICT
    if warn is None:
        warn = _WARN
    if return_all is None:
        return_all = _RETURN_ALL

    # dsig = {ids: {'sig': [...], 'comp': [...]}}
    dsig = _checkformat_getdata_dsig(
        dsig,
        occ=occ,
        indch=indch,
        indt=indt,
        dids=dids,
        dshort=dshort,
        dcomp=dcomp,
        dall_except=dall_except)

    # ------------------
    # get data

    anyfail = False
    dout = {ids: {} for ids in dsig.keys()}
    dfail = {ids: {} for ids in dsig.keys()}
    for ids in dsig.keys():
        indchi = dsig[ids]['indch']
        indti = dsig[ids]['indt']
        occi = dsig[ids]['occ']
        for ii in range(len(dsig[ids]['sig'])):
            sigi = dsig[ids]['sig'][ii]
            compi = bool(dsig[ids]['comp'][ii])
            if isclose is None:
                isclose_ = sigi == 't'
            else:
                isclose_ = isclose
            try:
                dout[ids][sigi] = _get_data_units(
                    ids, sigi, occi, comp=compi,
                    indt=indti, indch=indchi,
                    stack=stack, isclose=isclose_, flatocc=flatocc,
                    data=data, units=units,
                    nan=nan, pos=pos, empty=empty,
                    dids=dids, dcomp=dcomp, dshort=dshort,
                    dall_except=dall_except,
                    strict=strict,
                )

                lc = [dout[ids][sigi]['errdata'] is not None,
                      dout[ids][sigi]['errunits'] is not None]
                if any(lc):
                    anyfail = True
                    if lc[0]:
                        dfail[ids][sigi+' data'] = dout[ids][sigi]['errdata']
                    if lc[1]:
                        dfail[ids][sigi+' units'] = dout[ids][sigi]['errunits']

            except Exception as err:
                dfail[ids][sigi] = err
                anyfail = True

        if len(dfail[ids]) == 0:
            del dfail[ids]

    # ---------------------
    # Print if any failure

    if anyfail:

        if strict is True:
            for ids, vids in dfail.items():
                for sigi, verr in vids.items():
                    print(f"\n{ids}\t{sigi}\t{verr}")
                    raise verr

        if data is True:
            for ids in dfail.keys():
                for sigi in list(dout[ids].keys()):
                    if dout[ids][sigi]['errdata'] is not None:
                        del dout[ids][sigi]

        if warn:
            msg = "The following data could not be retrieved:"
            for ids in dfail.keys():
                nmax = np.max([len(k1) for k1 in dfail[ids].keys()])
                msg += "\n\t- {}:".format(ids)
                for sigi, erri in dfail[ids].items():
                    msgi = str(erri).replace('\n', ' ')
                    msg += f"\n\t\t{sigi.ljust(nmax)}:  {msgi}"
            warnings.warn(msg)

    # return
    if return_all:
        return dout, dfail, dsig
    else:
        return dout


# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import Polygon as plg
import datastock as ds


# specific
from . import _generic_check
from . import _generic_plot
from . import _class8_plot as _plot
from ..geom import _core
from . import _class8_plot_vos_spectro as _plot_vos_spectro


# ###############################################################
# ###############################################################
#                           plot main
# ###############################################################


def _plot_diagnostic_vos(
    coll=None,
    key=None,
    key_cam=None,
    optics=None,
    elements=None,
    proj=None,
    los_res=None,
    indch=None,
    indlamb=None,
    # data plot
    dvos=None,
    units=None,
    cmap=None,
    vmin=None,
    vmax=None,
    vmin_tot=None,
    vmax_tot=None,
    vmin_cam=None,
    vmax_cam=None,
    dvminmax=None,
    alpha=None,
    plot_colorbar=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
):

    # ------------
    # check inputs

    (
        key,
        key_cam,
        is2d,
        proj,
        ddata,
        dref,
        static,
        daxis,
        _,
        _,
        _,
        alpha,
        units,
        los_res,
        color_dict,
        _,
        _,
        _,
        ylab,
        plot_colorbar,
        _,
    ) = _plot._plot_diagnostic_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # parameters
        vmin=vmin,
        vmax=vmax,
        alpha=alpha,
        # figure
        plot_colorbar=plot_colorbar,
        proj=proj,
        units=units,
        los_res=los_res,
        # interactivity
        color_dict=color_dict,
    )

    # single camera + get dvos
    key_cam = key_cam[:1]
    key, dvos, isstore = coll.check_diagnostic_dvos(
        key,
        key_cam=key_cam,
        dvos=dvos,
    )
    doptics = coll.dobj['diagnostic'][key]['doptics'][key_cam[0]]
    shape_cam = coll.dobj['camera'][key_cam[0]]['dgeom']['shape']

    # indch
    if indch is None:
        if is2d:
            indch = [int(ss/2) for ss in shape_cam]
        else:
            indch = int(shape_cam[0]/2)

    # ------------
    # prepare data

    dplot = coll.get_diagnostic_dplot(
        key=key,
        key_cam=key_cam,
        optics=optics,
        elements=elements,
    )

    # ---------------------
    # prepare los and ddata

    if is2d:
        out0, out1 = coll.get_optics_outline(
            key=key_cam[0],
            add_points=False,
            total=True,
        )

        dgeom = coll.dobj['camera'][key_cam[0]]['dgeom']
        k0, k1 = dgeom['cents']
        x0 = coll.ddata[k0]['data']
        x1 = coll.ddata[k1]['data']
        if x0.size == 1:
            dx0 = coll.ddata[dgeom['outline'][0]]['data']
            dx0 = dx0.max() - dx0.min()
        else:
            dx0 = x0[1] - x0[0]
        if x1.size == 1:
            dx1 = coll.ddata[dgeom['outline'][1]]['data']
            dx1 = dx1.max() - dx1.min()
        else:
            dx1 = x1[1] - x1[0]
        extent_cam = (
            x0[0] - 0.5*dx0,
            x0[-1] + 0.5*dx0,
            x1[0] - 0.5*dx1,
            x1[-1] + 0.5*dx1,
        )
    else:
        x0, x1, extent_cam = None, None, None

    # ---------------------
    # prepare los and vos

    (
        los_x, los_y, los_z, los_r,
        los_xi, los_yi, los_zi, los_ri,
    ) = _prepare_los(
        coll=coll,
        doptics=doptics,
        los_res=los_res,
        is2d=is2d,
        indch=indch,
    )

    (
        pc0, pc1, ph0, ph1, pc0i, pc1i, ph0i, ph1i
    ) = _prepare_vos(
        coll=coll,
        doptics=doptics,
        los_res=los_res,
        is2d=is2d,
        indch=indch,
    )

    # mesh envelop
    dout = coll.get_mesh_outline(key=dvos[key_cam[0]]['keym'])
    p0 = dout['x0']['data']
    p1 = dout['x1']['data']

    # etendue and length
    etendue, length = _get_etendue_length(
        coll=coll,
        doptics=doptics,
        poly=np.array([p0, p1]),
    )

    # spectro => call routine
    spectro = coll.dobj['diagnostic'][key]['spectro']
    if spectro:
        return _plot_vos_spectro._plot(**locals())

    # dsamp from mesh
    sang_tot, sang_integ, sang, extent, sang_units = _prepare_sang(
        coll=coll,
        dvos=dvos[key_cam[0]],
        key_cam=key_cam,
        indch=indch,
        spectro=spectro,
        is2d=is2d,
    )

    # vmin, vmax
    if vmin is None:
        vmin = 0.
    if vmax is None:
        vmax = np.nanmax(sang)
    if vmin_tot is None:
        vmin_tot = 0.
    if vmax_tot is None:
        vmax_tot = np.nanmax(sang_tot)
    if vmin_cam is None:
        vmin_cam = 0.
    if vmax_cam is None:
        vmax_cam = np.nanmax(sang_integ)

    # -----------------
    # prepare figure

    if dax is None:

        dax = _get_dax(
            proj=proj,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
            is2d=is2d,
            key_cam=key_cam,
            indch=indch,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot diag elements

    for k0, v0 in dplot.items():

        for k1, v1 in v0.items():

            # cross
            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['r'],
                        v1['z'],
                        v1['ur'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['r'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['ux'],
                        v1['uy'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        **v1.get('props', {}),
                    )

    # ------------------
    # plot sang

    # crosstot
    kax = 'crosstot'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        im = ax.imshow(
            sang_tot.T,
            extent=extent,
            origin='lower',
            aspect='equal',
            interpolation='nearest',
            vmin=vmin_tot,
            vmax=vmax_tot,
        )

        ax.plot(
            np.r_[p0, p0[0]],
            np.r_[p1, p1[0]],
            c='k',
            lw=1.,
            ls='-',
        )

        if plot_colorbar is True:
            plt.colorbar(im, ax=ax, label=sang_units)

    # cross
    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        im = ax.imshow(
            sang.T,
            extent=extent,
            origin='lower',
            aspect='equal',
            interpolation='nearest',
            vmin=vmin,
            vmax=vmax,
        )

        if plot_colorbar is True:
            plt.colorbar(im, ax=ax, label=sang_units)

    # ------------------
    # plot los / vos

    # crosstot
    kax = 'crosstot'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _add_camera_los_cross(
            ax=ax,
            is2d=is2d,
            los_r=los_r,
            los_z=los_z,
            pc0=pc0,
            pc1=pc1,
            alpha=alpha,
            color_dict=color_dict,
        )

    # cross
    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _add_camera_los_cross(
            ax=ax,
            is2d=is2d,
            los_r=los_ri,
            los_z=los_zi,
            pc0=pc0i,
            pc1=pc1i,
            alpha=alpha,
            color_dict=color_dict,
        )

    # hor
    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _add_camera_los_hor(
            ax=ax,
            is2d=is2d,
            los_x=los_x,
            los_y=los_y,
            ph0=ph0,
            ph1=ph1,
            alpha=alpha,
            color_dict=color_dict,
        )

    # camera
    kax = key_cam[0]
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _add_camera_data(
            coll=coll,
            ax=ax,
            sang_integ=sang_integ,
            etendue=etendue,
            length=length,
            indch=indch,
            # vmin, vmax
            vmin_cam=vmin_cam,
            vmax_cam=vmax_cam,
            # 2d only
            is2d=is2d,
            ax_etend=dax.get(f'{key_cam[0]}_etend', {}).get('handle'),
            ax_diff=dax.get(f'{key_cam[0]}_diff', {}).get('handle'),
            x0=x0,
            x1=x1,
            extent_cam=extent_cam,
        )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

        kax = 'crosstot'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj='cross', dLeg=False)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

    # -------
    # connect

    return dax


# ##################################################################
# ##################################################################
#                       Prepare los and vos
# ##################################################################


def _prepare_los(
    coll=None,
    doptics=None,
    los_res=None,
    is2d=None,
    indch=None,
):

    # -----------
    # safety check

    if doptics['los'] is None:
        return [None] * 8

    # ----------
    # los

    los_x, los_y, los_z = coll.sample_rays(
        key=doptics['los'],
        res=los_res,
        mode='rel',
        concatenate=False,
    )

    # ------------------
    # for chosen index

    if is2d:
        los_xi = los_x[:, indch[0], indch[1]]
        los_yi = los_y[:, indch[0], indch[1]]
        los_zi = los_z[:, indch[0], indch[1]]
        los_ri = np.hypot(los_xi, los_yi)

        los_x, los_y, los_z, los_r = None, None, None, None

    else:
        los_r = np.hypot(los_x, los_y)

        los_xi = los_x[:, indch]
        los_yi = los_y[:, indch]
        los_zi = los_z[:, indch]
        los_ri = los_r[:, indch]

        # concatenate
        sh = tuple(np.r_[1, los_x.shape[1:]])
        los_x = np.append(los_x, np.full(sh, np.nan), axis=0).T.ravel()
        los_y = np.append(los_y, np.full(sh, np.nan), axis=0).T.ravel()
        los_z = np.append(los_z, np.full(sh, np.nan), axis=0).T.ravel()
        los_r = np.append(los_r, np.full(sh, np.nan), axis=0).T.ravel()

    return los_x, los_y, los_z, los_r, los_xi, los_yi, los_zi, los_ri


def _prepare_vos(
    coll=None,
    doptics=None,
    los_res=None,
    is2d=None,
    indch=None,
):

    # -----------
    # safety check

    if doptics['los'] is None:
        return [None] * 8

    # -------
    # vos

    kpc = doptics['dvos']['pcross']
    pc0 = coll.ddata[kpc[0]]['data']
    pc1 = coll.ddata[kpc[1]]['data']
    kph = doptics['dvos']['phor']
    ph0 = coll.ddata[kph[0]]['data']
    ph1 = coll.ddata[kph[1]]['data']

    if is2d:
        pc0i = pc0[:, indch[0], indch[1]]
        pc1i = pc1[:, indch[0], indch[1]]
        ph0i = ph0[:, indch[0], indch[1]]
        ph1i = ph1[:, indch[0], indch[1]]

        # envelop
        pc0 = pc0.reshape(pc0.shape[0], -1)
        pc1 = pc1.reshape(pc1.shape[0], -1)
        ph0 = ph0.reshape(ph0.shape[0], -1)
        ph1 = ph1.reshape(ph1.shape[0], -1)

        iok = np.all(np.isfinite(pc0), axis=0).nonzero()[0]
        pc = plg.Polygon(np.array([pc0[:, iok[0]], pc1[:, iok[0]]]).T)
        ph = plg.Polygon(np.array([ph0[:, iok[0]], ph1[:, iok[0]]]).T)
        for ii in iok[1:]:
            pc = pc | plg.Polygon(np.array([pc0[:, ii], pc1[:, ii]]).T)
            ph = ph | plg.Polygon(np.array([ph0[:, ii], ph1[:, ii]]).T)

        # -----------------------
        # convex hull if distinct

        # pc
        if len(pc) > 1:
            # replace by convex hull
            pts = np.concatenate(
                tuple([np.array(pc.contour(ii)) for ii in range(len(pc))]),
                axis=0,
            )
            pc0, pc1 = pts[ConvexHull(pts).vertices, :].T
        else:
            pc0, pc1 = np.array(pc).T

        # ph
        if len(ph) > 1:
            # replace by convex hull
            pts = np.concatenate(
                tuple([np.array(ph.contour(ii)) for ii in range(len(ph))]),
                axis=0,
            )
            ph0, ph1 = pts[ConvexHull(pts).vertices, :].T
        else:
            ph0, ph1 = np.array(ph).T

    else:
        pc0i = pc0[:, indch]
        pc1i = pc1[:, indch]
        ph0i = ph0[:, indch]
        ph1i = ph1[:, indch]

    # safety check
    if np.any(~np.isfinite(pc0i)):
        pc0i, pc1i = None, None
        ph0i, ph1i = None, None

    return pc0, pc1, ph0, ph1, pc0i, pc1i, ph0i, ph1i


# ################################################################
# ################################################################
#                       Prepare sang
# ################################################################


def _prepare_sang(
    coll=None,
    dvos=None,
    key_cam=None,
    indch=None,
    spectro=None,
    is2d=None,
):

    # -----------------
    # get mesh sampling

    dsamp = coll.get_sample_mesh(
        key=dvos['keym'],
        res=dvos['res_RZ'],
        mode='abs',
        grid=False,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    # -----------------
    # prepare image

    n0, n1 = dsamp['x0']['data'].size, dsamp['x1']['data'].size
    shape = (n0, n1)
    sang = np.full(shape, np.nan)
    sang_tot = np.full(shape, 0.)

    if is2d:
        for ii in range(dvos['indr_cross']['data'].shape[0]):
            for jj in range(dvos['indr_cross']['data'].shape[1]):
                iok = dvos['indr_cross']['data'][ii, jj, :] >= 0
                indr = dvos['indr_cross']['data'][ii, jj, iok]
                indz = dvos['indz_cross']['data'][ii, jj, iok]
                sang_tot[indr, indz] += dvos['sang_cross']['data'][ii, jj, iok]

                # sang
                if ii == indch[0] and jj == indch[1]:
                    sang[indr, indz] = dvos['sang_cross']['data'][ii, jj, iok]

    else:
        for ii in range(dvos['indr_cross']['data'].shape[0]):
            iok = dvos['indr_cross']['data'][ii, :] >= 0
            indr = dvos['indr_cross']['data'][ii, iok]
            indz = dvos['indz_cross']['data'][ii, iok]
            sang_tot[indr, indz] += dvos['sang_cross']['data'][ii, iok]

            # sang
            if ii == indch:
                sang[indr, indz] = dvos['sang_cross']['data'][ii, iok]

    sang_tot[sang_tot == 0.] = np.nan
    sang[sang == 0.] = np.nan

    # -------------------
    # get integrated vos

    sang_integ = np.nansum(dvos['sang_cross']['data'], axis=-1)

    # extent
    x0 = dsamp['x0']['data']
    dx0 = x0[1] - x0[0]
    x1 = dsamp['x1']['data']
    dx1 = x1[1] - x1[0]

    extent = (
        x0[0] - 0.5*dx0,
        x0[-1] + 0.5*dx0,
        x1[0] - 0.5*dx1,
        x1[-1] + 0.5*dx1,
    )

    # units
    sang_units = dvos['sang_cross']['units']

    return sang_tot, sang_integ, sang, extent, sang_units


def _get_etendue_length(
    coll=None,
    doptics=None,
    poly=None,
):

    # -----------
    # safety check

    if doptics['etendue'] is None:
        return None, None

    # ----------------------
    # get etendue and length

    ketend = doptics['etendue']
    etendue = coll.ddata[ketend]['data']

    # -------------------------
    # los through mesh envelopp

    # los
    klos = doptics['los']
    ptsx, ptsy, ptsz = coll.get_rays_pts(key=klos)
    vectx, vecty, vectz = coll.get_rays_vect(key=klos)

    iok = np.isfinite(vectx[-1, ...])
    length = np.full(vectx.shape[1:], np.nan)

    DD = np.array([
        ptsx[-2, ...][iok],
        ptsy[-2, ...][iok],
        ptsz[-2, ...][iok],
    ])
    uu = np.array([
        vectx[-1, ...][iok],
        vecty[-1, ...][iok],
        vectz[-1, ...][iok],
    ])


    # Prepare structures
    ves = _core.Ves(
        Poly=poly,
        Name='temp',
        Exp='',
    )

    conf = _core.Config(
        lStruct=[ves],
        Name='temp',
        Exp='',
    )

    # ray-tracing
    cam = _core.CamLOS1D(
        dgeom=(DD, uu),
        config=conf,
        Name='temp',
        Diag='',
        Exp='',
        strict=False,
    )

    # length
    length[iok] = (cam.dgeom['kOut'] - cam.dgeom['kIn'])

    return etendue, length


# ###############################################################
# ###############################################################
#                       add mobile
# ###############################################################


def _add_camera_los_cross(
    ax=None,
    is2d=None,
    color_dict=None,
    los_r=None,
    los_z=None,
    pc0=None,
    pc1=None,
    alpha=None,
):

    # ------
    # los

    if (not is2d) and (los_r is not None):
        l0, = ax.plot(
            los_r,
            los_z,
            c='k',
            ls='-',
            lw=1.,
        )

    # ------
    # vos
    if pc0 is not None:
        if pc0.ndim == 2:
            for ii in range(pc0.shape[1]):
                l0, = ax.fill(
                    pc0[:, ii],
                    pc1[:, ii],
                    fc='k',
                    alpha=alpha,
                    ls='None',
                    lw=0.,
                )

        elif pc0.ndim == 3:
            for ii in range(pc0.shape[1]):
                for jj in range(pc0.shape[2]):
                    l0, = ax.fill(
                        pc0[:, ii, jj],
                        pc1[:, ii, jj],
                        fc='k',
                        alpha=alpha,
                        ls='None',
                        lw=0.,
                    )

        else:
            l0, = ax.fill(
                pc0,
                pc1,
                fc='k',
                alpha=alpha,
                ls='None',
                lw=0.,
            )


def _add_camera_los_hor(
    ax=None,
    is2d=None,
    los_x=None,
    los_y=None,
    ph0=None,
    ph1=None,
    alpha=None,
    color_dict=None,
):

    # ------
    # los

    if (not is2d) and (los_x is not None):
        l0, = ax.plot(
            los_x,
            los_y,
            c=color_dict['x'][0],
            ls='-',
            lw=1.,
        )

    # ------
    # vos

    if ph0 is not None:

        if ph0.ndim == 2:
            for ii in range(ph0.shape[1]):
                l0, = ax.fill(
                    ph0[:, ii],
                    ph1[:, ii],
                    fc='k',
                    alpha=alpha,
                    ls='None',
                    lw=0.,
                )

        elif ph0.ndim == 3:
            for ii in range(ph0.shape[1]):
                for jj in range(ph0.shape[2]):
                    l0, = ax.fill(
                        ph0[:, ii, jj],
                        ph1[:, ii, jj],
                        fc='k',
                        alpha=alpha,
                        ls='None',
                        lw=0.,
                    )

        else:
            l0, = ax.fill(
                ph0,
                ph1,
                fc='k',
                alpha=alpha,
                ls='None',
                lw=0.,
            )


def _add_camera_data(
    coll=None,
    ax=None,
    sang_integ=None,
    etendue=None,
    length=None,
    indch=None,
    color_dict=None,
    # vmin, vmax
    vmin_cam=None,
    vmax_cam=None,
    # 2d only
    ax_diff=None,
    ax_etend=None,
    is2d=None,
    x0=None,
    x1=None,
    extent_cam=None,
):


    if is2d:

        # sang
        if sang_integ is not None:
            mi = ax.imshow(
                sang_integ.T,
                origin='lower',
                extent=extent_cam,
                interpolation='nearest',
                vmin=vmin_cam,
                vmax=vmax_cam,
            )

        # etendue
        if etendue is not None:
            etendle = etendue * length
            mi = ax_etend.imshow(
                etendle.T,
                origin='lower',
                extent=extent_cam,
                interpolation='nearest',
                vmin=vmin_cam,
                vmax=vmax_cam,
            )
            plt.colorbar(mi, ax=[ax, ax_etend])

        # diff
        if sang_integ is not None and etendue is not None:
            diff = (sang_integ - etendle).T
            dmax = np.abs(max(np.nanmin(diff), np.nanmax(diff)))
            imd = ax_diff.imshow(
                (sang_integ - etendle).T,
                origin='lower',
                extent=extent_cam,
                interpolation='nearest',
                cmap=plt.cm.seismic,
                vmin=-dmax,
                vmax=dmax,
            )

            plt.colorbar(imd, ax=ax_diff)

        # marker
        if x0 is not None:
            for aa in [ax, ax_etend, ax_diff]:
                aa.plot(
                    [x0[indch[0]]],
                    [x1[indch[1]]],
                    c='k',
                    marker='s',
                    ms=6,
                    ls='None',
                    lw=1.,
                )

    else:

        nlos = sang_integ.shape[0]
        ind = np.arange(0, nlos)

        # vos
        if sang_integ is not None:
            nlos = sang_integ.shape[0]
            ind = np.arange(0, nlos)
            ax.plot(
                ind,
                sang_integ,
                c='b',
                marker='.',
                ls='-',
                lw=1.,
                label="sang * dV (sr.m3)",
            )

        # los
        if etendue is not None:
            nlos = etendue.shape[0]
            ind = np.arange(0, nlos)
            ax.plot(
                ind,
                etendue * length,
                c='k',
                marker='.',
                ls='-',
                lw=1.,
                label="etendue * length (sr.m3)",
            )

        # indch
        ax.axvline(
            indch,
            c='k',
            lw=1.,
            ls='--',
        )

        ax.set_ylim(vmin_cam, vmax_cam)
        plt.legend()


# ################################################################
# ################################################################
#                   figure
# ################################################################


def _get_dax(
    proj=None,
    dmargin=None,
    fs=None,
    tit=None,
    wintit=None,
    is2d=None,
    key_cam=None,
    indch=None,
):

    # ------------
    # check inputs

    # fs
    if fs is None:
        fs = (14.5, 10)

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.05, 'top': 0.95,
            'left': 0.05, 'right': 0.98,
            'wspace': 0.20, 'hspace': 0.40,
            # 'width_ratios': [0.6, 0.4],
            # 'height_ratios': [0.4, 0.6],
        },
    )

    # wintit
    wintit = ds._generic_check._check_var(
        wintit, 'wintit',
        types=str,
        default=_generic_plot._WINDEF,
    )

    # tit
    if tit is None:
        tit = f"{key_cam} - {indch}\nVOS"

    # -------------
    # Create figure

    fig = plt.figure(figsize=fs)
    fig.canvas.manager.set_window_title(wintit)
    fig.suptitle(tit, size=12, fontweight='bold')

    if is2d:

        gs = gridspec.GridSpec(ncols=4, nrows=4, **dmargin)

        # -------------
        # hor

        ax0 = fig.add_subplot(gs[2:, 2:], aspect='equal', adjustable='datalim')
        ax0.set_xlabel(r'X (m)', size=12)
        ax0.set_ylabel(r'Y (m)', size=12)

        # -------------
        # cross

        ax1 = fig.add_subplot(gs[:2, :2], aspect='equal')  # , adjustable='datalim')
        ax1.set_xlabel(r'R (m)', size=12)
        ax1.set_ylabel(r'Z (m)', size=12)

        # -------------
        # cross

        ax2 = fig.add_subplot(gs[2:, :2], sharex=ax1, sharey=ax1)
        ax2.set_xlabel(r'R (m)', size=12)
        ax2.set_ylabel(r'Z (m)', size=12)

        # -------------
        # camera

        ax3 = fig.add_subplot(gs[0, 2], aspect='equal')  # , adjustable='datalim')
        ax3.set_xlabel(r'x0 (m)', size=12)
        ax3.set_ylabel(r'x1 (m)', size=12)
        ax3.set_title('sang * dV (m3.sr)', size=12, fontweight='bold')

        ax4 = fig.add_subplot(gs[0, 3], sharex=ax3, sharey=ax3)
        ax4.set_xlabel(r'x0 (m)', size=12)
        ax4.set_title('etendue * length (m3.sr)', size=12, fontweight='bold')

        ax5 = fig.add_subplot(gs[1, 2], sharex=ax3, sharey=ax3)
        ax5.set_xlabel(r'x0 (m)', size=12)
        ax5.set_ylabel(r'x1 (m)', size=12)
        ax5.set_title('difference (m3.sr)', size=12, fontweight='bold')

        # ---------
        # dict

        dax = {
            'hor': {'handle': ax0, 'type': 'hor'},
            'cross': {'handle': ax1, 'type': 'cross'},
            'crosstot': {'handle': ax2, 'type': 'cross'},
            key_cam[0]: {'handle': ax3, 'type': 'camera'},
            f'{key_cam[0]}_etend': {'handle': ax4, 'type': 'camera'},
            f'{key_cam[0]}_diff': {'handle': ax5, 'type': 'camera'},
        }

    else:

        gs = gridspec.GridSpec(ncols=2, nrows=2, **dmargin)

        # -------------
        # hor

        ax0 = fig.add_subplot(gs[1, 1], aspect='equal', adjustable='datalim')
        ax0.set_xlabel(r'X (m)', size=12)
        ax0.set_ylabel(r'Y (m)', size=12)

        # -------------
        # cross

        ax1 = fig.add_subplot(gs[0, 0], aspect='equal')  # adjustable='datalim')
        ax1.set_xlabel(r'R (m)', size=12)
        ax1.set_ylabel(r'Z (m)', size=12)

        # -------------
        # cross

        ax2 = fig.add_subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        ax2.set_xlabel(r'R (m)', size=12)
        ax2.set_ylabel(r'Z (m)', size=12)

        # -------------
        # camera

        ax3 = fig.add_subplot(gs[0, 1])
        ax3.set_xlabel(r'index', size=12)
        ax3.set_ylabel(r'data', size=12)

        # ---------
        # dict

        dax = {
            'hor': {'handle': ax0, 'type': 'hor'},
            'cross': {'handle': ax1, 'type': 'cross'},
            'crosstot': {'handle': ax2, 'type': 'cross'},
            key_cam[0]: {'handle': ax3, 'type': 'camera'},
        }

    return dax


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: initializedcheck=False
#
################################################################################
# Utility functions for basic geometry :
#   - vector calculus (cross product, dot product, norm, ...)
#   - cythonization of matplotlib path functions (is point in a path?)
#   - cythonization of some numpy functions (hypotenus, tile, sum)
################################################################################
from libc.stdint cimport int64_t
cimport cython
from cpython.array cimport array, clone

# ==============================================================================
# ==  Geometry global variables
# ==============================================================================
# Values defined in the *.pyx file
cdef double _VSMALL
cdef double _SMALL
cdef double _TWOPI

# ==============================================================================
# == Redifinition of functions
# ==============================================================================
cdef bint is_point_in_path(const int nvert,
                           const double* vertx,
                           const double* verty,
                           const double testx,
                           const double testy) nogil

cdef int is_point_in_path_vec(const int nvert,
                              const double* vertx,
                              const double* verty,
                              const int npts,
                              const double* testx,
                              const double* testy,
                              int* is_in_path) nogil

cdef void compute_inv_and_sign(const double[3] ray_vdir,
                               int[3] sign,
                               double[3] inv_direction) nogil

cdef array compute_hypot(const double[::1] xpts, const double[::1] ypts,
                         int npts=*)

cdef double comp_min_hypot(const double[::1] xpts, const double[::1] ypts,
                           int npts=*) nogil

cdef double comp_min(double[::1] vec, int npts) nogil

cdef void tile_3_to_2d(double v0, double v1, double v2, int npts,
                       double[:,::1] res) nogil
# ==============================================================================
# =  Polygon helpers
# ==============================================================================
cdef int find_ind_lowerright_corner(const double[::1] xpts,
                                    const double[::1] ypts,
                                    int npts) nogil

# ==============================================================================
# == Vector Calculus Helpers
# ==============================================================================
cdef void compute_cross_prod(const double[3] vec_a,
                             const double[3] vec_b,
                             double[3] res) nogil

cdef void compute_dist_pt_vec(const double pt0, const double pt1,
                              const double pt2, const int npts,
                              const double[:, ::1] vec,
                              double* dist) nogil

cdef void compute_dist_vec_vec(const int npts1, const int npts2,
                               const double[:, ::1] vec1,
                               const double[:, ::1] vec2,
                               double[:, ::1] dist) nogil

cdef double compute_norm(const double[3] vec) nogil

cdef  double compute_dot_prod(const double[3] vec_a,
                              const double[3] vec_b) nogil

cdef  double compute_g(double s, double m2b2, double rm0sqr,
                       double m0sqr, double b1sqr) nogil

cdef  double compute_bisect(double m2b2, double rm0sqr,
                            double m0sqr, double b1sqr,
                            double smin, double smax) nogil

cdef  double compute_find(double m2b2, double rm0sqr,
                          double m0sqr, double b1sqr,
                          double t0, double t1, double f0, double f1,
                          int maxIterations, double root) nogil

cdef void compute_diff_div(const double[:, ::1] vec1,
                           const double[:, ::1] vec2,
                           const double* div,
                           const int npts,
                           double[:, ::1] res) nogil

# ==============================================================================
# == Matrix sum (np.sum)
# ==============================================================================
cdef int64_t sum_naive_int(int64_t* orig, int n_cols) nogil


"""
This module contains tests for tofu.geom solid angle routines
"""


# Built-in
import os
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt

# tofu-specific
from ... import geom as tfg


_PATH_HERE = os.path.dirname(__file__)
_PATH_OUTPUT = os.path.join(_PATH_HERE, 'output')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def clean(path=_PATH_OUTPUT):
    pass


def setup_module(module):
    clean()


def teardown_module(module):
    clean()


#######################################################
#
#     Utilities
#
#######################################################


def _create_poly_2d_ccw():
    """ Reference case, example taken from:

    https://www.geometrictools.com/Documentation/TriangulationByEarClipping.pdf
    """
    # add references (i.e.: store size of each dimension under a unique key)
    outline_x0 = np.r_[0, 10, 20, 30, 40, 28, 22, 15, 4, 6]
    outline_x1 = np.r_[0, -20, 5, -10, 20, 15, 40, 0, 10, 30]

    reflex0 = np.r_[2, 5, 7, 8]
    ears0 = np.r_[3, 4, 6, 9]

    tri = np.array([
        [2, 3, 4],
        [2, 4, 5],
        [2, 5, 6],
        [2, 6, 7],
        [1, 2, 7],
        [0, 1, 7],
        [0, 7, 8],
        [0, 8, 9],
    ])

    cents = np.r_[10, 0, 0]

    nin = np.r_[-1, 0, 0]
    e0 = np.r_[0, -1, 0]
    e1 = np.r_[0, 0, 1]

    return {
        'outline_x0': outline_x0,
        'outline_x1': outline_x1,
        'reflex0': reflex0,
        'ears0': ears0,
        'tri': tri,
        'cents_x': cents[0],
        'cents_y': cents[1],
        'cents_z': cents[2],
        'nin_x': nin[0],
        'nin_y': nin[1],
        'nin_z': nin[2],
        'e0_x': e0[0],
        'e0_y': e0[1],
        'e0_z': e0[2],
        'e1_x': e1[0],
        'e1_y': e1[1],
        'e1_z': e1[2],
    }


def _create_single_triangle():

    poly_x = np.r_[1, 0, 0]
    poly_y = np.r_[0, 1, 0]
    poly_z = np.r_[0, 0, 1]

    cents_x = 1/3.
    cents_y = 1/3.
    cents_z = 1/3.

    u = np.r_[1, 1, 1]
    nin = -u / np.sqrt(3)
    e0 = np.r_[nin[1], -nin[0], 0.]
    e0 = e0 / np.linalg.norm(e0)
    e1 = np.cross(nin, e0)

    outx0 = (
        (poly_x - cents_x)*e0[0]
        + (poly_y - cents_y)*e0[1]
        + (poly_z - cents_z)*e0[2]
    )
    outx1 = (
        (poly_x - cents_x)*e1[0]
        + (poly_y - cents_y)*e1[1]
        + (poly_z - cents_z)*e1[2]
    )

    pts_x = np.r_[0, 1] * u[0]
    pts_y = np.r_[0, 1] * u[1]
    pts_z = np.r_[0, 1] * u[2]

    sa = (4*np.pi / 8.) * np.r_[1, 0]

    return {
        'poly_x': poly_x,
        'poly_y': poly_y,
        'poly_z': poly_z,
        'outline_x0': outx0,
        'outline_x1': outx1,
        'cents_x': cents_x,
        'cents_y': cents_y,
        'cents_z': cents_z,
        'nin_x': nin[0],
        'nin_y': nin[1],
        'nin_z': nin[2],
        'e0_x': e0[0],
        'e0_y': e0[1],
        'e0_z': e0[2],
        'e1_x': e1[0],
        'e1_y': e1[1],
        'e1_z': e1[2],
        'pts_x': pts_x,
        'pts_y': pts_y,
        'pts_z': pts_z,
        'sa': sa,
    }


def _create_single_rectangle():

    a, b = 2, 1
    outline_x0 = a*np.r_[-1, 1, 1, -1]
    outline_x1 = b*np.r_[-1, -1, 1, 1]

    cents = np.r_[0, 0, 0]

    nin = np.r_[0, 0, 1]
    e0 = np.r_[1, 0, 0]
    e1 = np.r_[0, 1, 0]

    pts_z = np.r_[-1, 0.001, 0.1, 0.5, 1, 10, 50]
    pts_x = np.zeros((pts_z.size,))
    pts_y = np.zeros((pts_z.size,))

    # For a rectangle of dimensions (a, b) seen from height h
    # https://www.planetmath.org/solidangleofrectangularpyramid
    # sa = 4 arcsin( ab / sqrt( (a^2 + h^2)*(b^2 + h^2) ) )
    h = pts_z[1:]
    sa = 4*np.arcsin(a*b / np.sqrt((a**2 + h**2) * (b**2 + h**2)))
    sa = np.r_[0, sa]

    return {
        'outline_x0': outline_x0,
        'outline_x1': outline_x1,
        'cents_x': cents[0],
        'cents_y': cents[1],
        'cents_z': cents[2],
        'nin_x': nin[0],
        'nin_y': nin[1],
        'nin_z': nin[2],
        'e0_x': e0[0],
        'e0_y': e0[1],
        'e0_z': e0[2],
        'e1_x': e1[0],
        'e1_y': e1[1],
        'e1_z': e1[2],
        'pts_x': pts_x,
        'pts_y': pts_y,
        'pts_z': pts_z,
        'sa': sa,
    }


def _create_light(npts=5):

    a, b = 2, 1
    outline_x0 = a*np.r_[-1, 1, 1, -1]
    outline_x1 = b*np.r_[-1, -1, 1, 1]

    cents = np.r_[0, 0, 0]

    nin = np.r_[0, 0, 1]
    e0 = np.r_[1, 0, 0]
    e1 = np.r_[0, 1, 0]

    det = {
        'outline_x0': outline_x0,
        'outline_x1': outline_x1,
        'cents_x': cents[0],
        'cents_y': cents[1],
        'cents_z': cents[2],
        'nin_x': nin[0],
        'nin_y': nin[1],
        'nin_z': nin[2],
        'e0_x': e0[0],
        'e0_y': e0[1],
        'e0_z': e0[2],
        'e1_x': e1[0],
        'e1_y': e1[1],
        'e1_z': e1[2],
    }

    ap = {
        'ap0': {
            'poly_x': 10*a*np.r_[-1, 1, 1, -1],
            'poly_y': 10*b*np.r_[-1, -1, 1, 1],
            'poly_z': 0.1*np.ones((4,)),
            'nin': nin,
        },
        'ap1': {
            'poly_x': 0.1*a*np.r_[-1, 1, 1, -1],
            'poly_y': 0.1*b*np.r_[-1, -1, 1, 1],
            'poly_z': 0.2*np.ones((4,)),
            'nin': nin,
        },
        'ap3': {
            'poly_x': 10*a*np.r_[-1, 0, 0, -1],
            'poly_y': 10*b*np.r_[-1, -1, 1, 1],
            'poly_z': 0.25*np.ones((4,)),
            'nin': nin,
        },
    }
    ap['ap2'] = {k0: dict(ap[k0]) for k0 in ['ap0', 'ap1']}
    ap['ap3'] = {k0: dict(ap[k0]) for k0 in ['ap0', 'ap1', 'ap3']}

    pts_z = np.r_[-1, -0.5, np.linspace(1, 50, npts)]
    pts_x = np.zeros((pts_z.size,))
    pts_y = np.zeros((pts_z.size,))

    # For a rectangle of dimensions (a, b) seen from height h
    # https://www.planetmath.org/solidangleofrectangularpyramid
    # sa = 4 arcsin( ab / sqrt( (a^2 + h^2)*(b^2 + h^2) ) )
    h = pts_z[2:]
    sa0 = 4*np.arcsin(a*b / np.sqrt((a**2 + h**2) * (b**2 + h**2)))

    h = pts_z[2:] - 0.2
    sa1 = 4*np.arcsin(
        0.01*a*b / np.sqrt((0.01*a**2 + h**2) * (0.01*b**2 + h**2))
    )

    h = pts_z[2:] - 0.2
    sa2 = 4*np.arcsin(
        0.01*a*b / np.sqrt((0.01*a**2 + h**2) * (0.01*b**2 + h**2))
    )
    sa0 = np.r_[0, 0, sa0]
    sa1 = np.r_[0, 0, sa1]
    sa2 = np.r_[0, 0, sa2]
    sa3 = sa2/2.

    uvx = np.r_[np.nan, np.nan, np.zeros((sa0.size-2,))]
    uvy = np.r_[np.nan, np.nan, np.zeros((sa0.size-2,))]
    uvz = np.r_[np.nan, np.nan, -np.ones((sa0.size-2,))]

    return {
        'pts_x': pts_x,
        'pts_y': pts_y,
        'pts_z': pts_z,
        'det': det,
        'ap': ap,
        'sa0': sa0,
        'sa1': sa1,
        'sa2': sa2,
        'sa3': sa3,
        'uvx': uvx,
        'uvy': uvy,
        'uvz': uvz,
    }


def _create_visibility(npts=4):

    conf = tfg.utils.create_config('ITER-V0')

    a, b = 2, 1
    outline_x0 = a*np.r_[-1, 1, 1, -1]
    outline_x1 = b*np.r_[-1, -1, 1, 1]

    cents = np.r_[5.6, 0, -3.8]

    nin = np.r_[-1, 0, 0]
    e0 = np.r_[0, 1, 0]
    e1 = np.cross(nin, e0)

    det = {
        'outline_x0': outline_x0,
        'outline_x1': outline_x1,
        'cents_x': cents[0],
        'cents_y': cents[1],
        'cents_z': cents[2],
        'nin_x': nin[0],
        'nin_y': nin[1],
        'nin_z': nin[2],
        'e0_x': e0[0],
        'e0_y': e0[1],
        'e0_z': e0[2],
        'e1_x': e1[0],
        'e1_y': e1[1],
        'e1_z': e1[2],
    }

    ap = {
        'ap0': {
            'poly_x': (cents[0] - 0.05)*np.ones((4,)),
            'poly_y': cents[1] + 10*a*np.r_[-1, 1, 1, -1],
            'poly_z': cents[2] + 10*b*np.r_[-1, -1, 1, 1],
            'nin': nin,
        },
        'ap1': {
            'poly_x': (cents[0] - 0.1)*np.ones((4,)),
            'poly_y': cents[1] + 0.1*a*np.r_[-1, 1, 1, -1],
            'poly_z': cents[2] + 0.1*b*np.r_[-1, -1, 1, 1],
            'nin': nin,
        },
        'ap3': {
            'poly_x': (cents[0] - 0.15)*np.ones((4,)),
            'poly_y': cents[1] + 10*a*np.r_[-1, 0, 0, -1],
            'poly_z': cents[2] + 10*b*np.r_[-1, -1, 1, 1],
            'nin': nin,
        },
    }

    # pts
    if npts == 4:
        pts_x = np.r_[5.4, 5.2, 4.8, 4.4]
    else:
        pts_x = np.linspace(5.4, 4.4, npts)
    pts_y = np.zeros((pts_x.size,))
    pts_z = cents[2] * np.ones((pts_x.size,))

    # sa
    h = cents[0] - pts_x - 0.1
    sa = 4*np.arcsin(
        0.01*a*b / np.sqrt((0.01*a**2 + h**2) * (0.01*b**2 + h**2))
    )
    sa = np.r_[sa[:2]*0.5, 0., 0.]

    return {
        'pts_x': pts_x,
        'pts_y': pts_y,
        'pts_z': pts_z,
        'det': det,
        'ap': ap,
        'sa': sa,
        'config': conf,
    }


def _create_etendue(res=None):

    # unit vectors
    nout, ei, ej = np.r_[-1, 0, 0], np.r_[0, -1, 0], np.r_[0, 0, 1]
    nout2 = nout*np.cos(np.pi/6) + ei*np.sin(np.pi/6.)
    ei2 = nout*np.sin(np.pi/6) - ei*np.cos(np.pi/6.)
    nout3 = nout*np.cos(-2.*np.pi/6) + ei*np.sin(-2.*np.pi/6.)
    ei3 = nout*np.sin(-2.*np.pi/6) - ei*np.cos(-2.*np.pi/6.)

    # outlines
    out0, out1 = np.r_[-1, 1, 1, -1, -1], np.r_[-1, -1, 1, 1, -1]
    outa0, outa1 = 2e-2*out0, 5e-3*out1
    outd0, outd1 = 0.04*out0, 0.04*out1

    # centers
    ca, cd = np.r_[10, 0, 0], np.r_[20, 0, 0]

    # aperture polygon
    poly_x = ca[0] + outa0*ei2[0] + outa1*ej[0]
    poly_y = ca[1] + outa0*ei2[1] + outa1*ej[1]
    poly_z = ca[2] + outa0*ei2[2] + outa1*ej[2]

    # aperture dict
    aperture = {
        'a0': {
            'poly_x': poly_x,
            'poly_y': poly_y,
            'poly_z': poly_z,
            'cent': ca,
            'nin': nout2,
            'e0': ei2,
            'e1': ej,
        },
    }

    # detector dict
    det = {
        'outline_x0': outd0,
        'outline_x1': outd1,
        'cents_x': cd[0],
        'cents_y': cd[1],
        'cents_z': cd[2],
        'nin_x': nout3[0],
        'nin_y': nout3[1],
        'nin_z': nout3[2],
        'e0_x': ei3[0],
        'e0_y': ei3[1],
        'e0_z': ei3[2],
        'e1_x': ej[0],
        'e1_y': ej[1],
        'e1_z': ej[2],
    }

    return {
        'aperture': aperture,
        'det': det,
        'etendue': np.r_[1.10852804e-08],
        'res': res,
    }


#######################################################
#
#     Instanciate
#
#######################################################


class Test01_SolidAngles():

    def setup_method(self):
        self.poly_2d_ccw = _create_poly_2d_ccw()
        self.single_triangle = _create_single_triangle()
        self.single_rectangle = _create_single_rectangle()
        self.light = _create_light()
        self.visibility = _create_visibility()
        self.etendue = _create_etendue()

    # ------------------------
    #   Populating
    # ------------------------

    def test01_triangulation_2d_rectangle(self):
        tri = tfg._comp_solidangles.triangulate_polygon_2d(
            2*np.r_[-1., 1, 1, -1],
            np.r_[-1., -1, 1, 1],
        )

        triref = np.array([[0, 1, 2], [1, 2, 3]])
        if np.allclose(tri, triref):
            msg = (
                "Wrong rectangle triangulation:\n"
                f"\t- expected: {triref}\n"
                f"\t- obtained: {tri}\n"
            )

    def test02_triangulation_2d_ccw(self):
        tri = tfg._comp_solidangles.triangulate_polygon_2d(
            self.poly_2d_ccw['outline_x0'],
            self.poly_2d_ccw['outline_x1'],
        )

        if not tri.shape == self.poly_2d_ccw['tri'].shape:
            msg = (
                "Wrong shape of triangulation:\n"
                f"\t- Expected: {self.poly_2d_ccw['tri'].shape}\n"
                f"\t- obtained: {tri.shape}\n"
            )
            raise Exception(msg)

        if not np.allclose(tri, self.poly_2d_ccw['tri']):
            msg = (
                "Wrong tringulation of reference test case!\n"
                f"\t- expected: {self.poly_2d_ccw['tri']}\n"
                f"\t- obtained: {tri}\n"
            )
            raise Exception(msg)

    def test03_solid_angle_triangle(self):
        # single triangle
        sa = tfg.calc_solidangle_apertures(
            pts_x=self.single_triangle['pts_x'],
            pts_y=self.single_triangle['pts_y'],
            pts_z=self.single_triangle['pts_z'],
            apertures=None,
            detectors=self.single_triangle,
            visibility=False,
            return_vector=False,
        ).ravel()

        # check solid angle value
        if np.any(sa < 0.):
            msg = (
                "Solid angle of triangle is negative!\n"
                f"\t- obtained: {sa}"
            )
            raise Exception(msg)

        saref = self.single_triangle['sa']
        if np.any(np.abs(sa - saref) > 1.e-10 * saref):
            msg = (
                "Solid angle of triangle is wrong!\n"
                f"\t- Expected: {saref}\n"
                f"\t- Obtained: {sa}\n"
            )
            raise Exception(msg)

    def test04_solid_angle_rectangle(self):
        # rectangle
        sa = tfg.calc_solidangle_apertures(
            pts_x=self.single_rectangle['pts_x'],
            pts_y=self.single_rectangle['pts_y'],
            pts_z=self.single_rectangle['pts_z'],
            apertures=None,
            detectors=self.single_rectangle,
            visibility=False,
            return_vector=False,
        ).ravel()

        # check solid angle value
        if np.any(sa < 0.):
            msg = (
                "Solid angle of triangle is negative!\n"
                f"\t- obtained: {sa}"
            )
            raise Exception(msg)

        saref = self.single_rectangle['sa']
        if np.any(np.abs(sa - saref) > 1.e-10 * saref):
            msg = (
                "Solid angle of triangle is wrong!\n"
                f"\t- Expected: {saref}\n"
                f"\t- Obtained: {sa}\n"
            )
            raise Exception(msg)

    def test05_solid_angle_light(self):

        for ii in range(4):
            sa = tfg.calc_solidangle_apertures(
                pts_x=self.light['pts_x'],
                pts_y=self.light['pts_y'],
                pts_z=self.light['pts_z'],
                apertures=self.light['ap'][f'ap{ii}'],
                detectors=self.light['det'],
                visibility=False,
                return_vector=False,
            ).ravel()

            # check solid angle value
            saref = self.light[f'sa{ii}']
            if np.any(np.abs(sa - saref) > 1.e-6 * saref):
                msg = (
                    f"Solid angle of light (case{ii}) is wrong!\n"
                    f"\t- Expected: {saref}\n"
                    f"\t- Obtained: {sa}\n"
                )
                raise Exception(msg)

    def test06_solid_angle_vector(self):
        for ii in range(3):
            sa, uvx, uvy, uvz = tfg.calc_solidangle_apertures(
                pts_x=self.light['pts_x'],
                pts_y=self.light['pts_y'],
                pts_z=self.light['pts_z'],
                apertures=self.light['ap'][f'ap{ii}'],
                detectors=self.light['det'],
                visibility=False,
                return_vector=True,
            )
            sa = sa.ravel()
            uvx = uvx.ravel()
            uvy = uvy.ravel()
            uvz = uvz.ravel()

            # check solid angle value
            saref = self.light[f'sa{ii}']
            if np.any(np.abs(sa - saref) > 1.e-6 * saref):
                msg = (
                    f"Solid angle of light (case{ii}) is wrong!\n"
                    f"\t- Expected: {saref}\n"
                    f"\t- Obtained: {sa}\n"
                )
                raise Exception(msg)

            # check unit vectors
            uvxr = self.light['uvx']
            uvyr = self.light['uvy']
            uvzr = self.light['uvz']
            c0 = (
                np.allclose(uvx, uvxr, equal_nan=True)
                and np.allclose(uvx, uvxr, equal_nan=True)
                and np.allclose(uvx, uvxr, equal_nan=True)
            )
            if not c0:
                msg = (
                    f"Unit vectors of light (case{ii}) are wrong!\n"
                    "\t- uvx:\n"
                    f"\t\t- expected {uvxr}\n"
                    f"\t\t- obtained {uvx}\n"
                    "\t- uvy:\n"
                    f"\t\t- expected {uvyr}\n"
                    f"\t\t- obtained {uvy}\n"
                    "\t- uvz:\n"
                    f"\t\t- expected {uvzr}\n"
                    f"\t\t- obtained {uvz}\n"
                )
                raise Exception(msg)

    def test07_solid_angle_visible(self):
        sa = tfg.calc_solidangle_apertures(
            pts_x=self.visibility['pts_x'],
            pts_y=self.visibility['pts_y'],
            pts_z=self.visibility['pts_z'],
            apertures=self.visibility['ap'],
            detectors=self.visibility['det'],
            config=self.visibility['config'],
            visibility=True,
            return_vector=False,
        ).ravel()

        # check solid angle value
        saref = self.visibility['sa']
        if np.any(np.abs(sa - saref) > 1.e-6 * saref):
            msg = (
                f"Solid angle of visibility is wrong!\n"
                f"\t- Expected: {saref}\n"
                f"\t- Obtained: {sa}\n"
            )
            raise Exception(msg)

    def test08_etendue(self):

        detend = tfg.compute_etendue(
            det=self.etendue['det'],
            aperture=self.etendue['aperture'],
            res=self.etendue['res'],
            check=True,
            verb=True,
        )

        # non-regression
        c0 = np.allclose(
            detend['numerical'][-1, 0],
            self.etendue['etendue'],
            atol=1e-11,
            rtol=1e-2,
        )
        if not c0:
            msg = "Regression detected!"
            raise Exception(msg)

        # check match
        c0 = np.allclose(
            detend['analytical'][-1, :],
            detend['numerical'][-1, :],
            atol=1e-11,
            rtol=1e-2,
        )

        if not c0:
            msg = (
                "Mismatching analytical vs numerical etendue!\n"
                f"\t- analytical: {detend['analytical'][-1, :]}\n"
                f"\t- numerical: {detend['numerical'][-1, :]}\n"
            )
            raise Exception(msg)

        # without check
        detend1 = tfg.compute_etendue(
            det=self.etendue['det'],
            aperture=self.etendue['aperture'],
            res=self.etendue['res'],
            check=False,
            verb=False,
            analytical=False,
            plot=False,
        )

        # check match
        c0 = np.allclose(
            detend['numerical'],
            detend1['numerical'],
        )
        if not c0:
            msg = (
                "Mismatching numerical check vs no_check etendue!\n"
                f"\t- with check: {detend['numerical']}\n"
                f"\t- w/o check:  {detend1['numerical']}\n"
            )
            raise Exception(msg)

        plt.close('all')



# -*- coding: utf-8 -*-


import itertools as itt


import datetime as dtm      # DB
import numpy as np
import scipy.interpolate as scpinterp
import scipy.stats as scpstats
from matplotlib.path import Path
import matplotlib.pyplot as plt
# import datastock as ds


from ..geom import _comp_solidangles
from . import _class8_vos_utilities as _utilities


# ###########################################################
# ###########################################################
#               Main
# ###########################################################


def _vos(
    # ressources
    coll=None,
    doptics=None,
    key_cam=None,
    dsamp=None,
    # inputs
    x0u=None,
    x1u=None,
    x0f=None,
    x1f=None,
    x0l=None,
    x1l=None,
    dx0=None,
    dx1=None,
    sh=None,
    res_RZ=None,
    res_phi=None,
    bool_cross=None,
    # user-defined limits
    user_limits=None,
    # keep
    keep3d=None,
    return_vector=None,
    # parameters
    margin_poly=None,
    config=None,
    visibility=None,
    verb=None,
    # debug
    debug=None,
    # timing
    timing=None,
    dt11=None,
    dt111=None,
    dt1111=None,
    dt2222=None,
    dt3333=None,
    dt4444=None,
    dt222=None,
    dt333=None,
    dt22=None,
    # unused
    **kwdargs,
):
    """ vos computation for broadband """

    # ---------------
    # prepare polygon

    if timing:
        t00 = dtm.datetime.now()     # DB

    # ----------------
    # user-defined vos
    # ----------------

    if user_limits is not None:

        if user_limits.get('pcross_user') is not None:
            xx, yy, zz, dind, ir, iz, iphi, dV = _vos_points(
                # polygons
                pcross0=user_limits['pcross_user'][0, :],
                pcross1=user_limits['pcross_user'][1, :],
                phor0=user_limits['phor_user'][0, :],
                phor1=user_limits['phor_user'][1, :],
                margin_poly=margin_poly,
                dphi=np.r_[-np.pi, np.pi],
                # sampling
                dsamp=dsamp,
                x0f=x0f,
                x1f=x1f,
                x0u=x0u,
                x1u=x1u,
                res=res_phi,
                dx0=dx0,
                dx1=dx1,
                # shape
                sh=sh,
            )

            shape = coll.dobj['camera'][key_cam]['dgeom']['shape']
            shape = np.r_[0, shape]

        elif user_limits.get('Dphi') is not None:
            # get temporary vos
            kpc0, kpc1 = doptics[key_cam]['dvos']['pcross']
            shape = coll.ddata[kpc0]['data'].shape
            pcross0 = coll.ddata[kpc0]['data']
            pcross1 = coll.ddata[kpc1]['data']

            # phor
            phor0 = user_limits['phor0'][key_cam]
            phor1 = user_limits['phor1'][key_cam]
            dphi = user_limits['dphi'][key_cam]

        else:
            msg = (
                "Something weird with pcross0:\n"
                f"user_limits: {user_limits}\n"
            )
            raise Exception(msg)

    else:

        # get temporary vos
        kpc0, kpc1 = doptics[key_cam]['dvos']['pcross']
        shape = coll.ddata[kpc0]['data'].shape
        pcross0 = coll.ddata[kpc0]['data']
        pcross1 = coll.ddata[kpc1]['data']

        # phor
        kph0, kph1 = doptics[key_cam]['dvos']['phor']
        phor0 = coll.ddata[kph0]['data']
        phor1 = coll.ddata[kph1]['data']
        dphi = doptics[key_cam]['dvos']['dphi']

    # pinhole?
    pinhole = doptics[key_cam]['pinhole']

    # ---------------
    # prepare det
    # ---------------

    dgeom = coll.dobj['camera'][key_cam]['dgeom']
    par = dgeom['parallel']
    is2d = dgeom['nd'] == '2d'
    cx, cy, cz = coll.get_camera_cents_xyz(key=key_cam)
    dvect = coll.get_camera_unit_vectors(key=key_cam)
    outline = dgeom['outline']
    out0 = coll.ddata[outline[0]]['data']
    out1 = coll.ddata[outline[1]]['data']

    # -----------
    # prepare lap
    # -----------

    optics = doptics[key_cam]['optics']
    dap0 = coll.get_optics_as_input_solid_angle(keys=optics)
    if pinhole is False:
        paths = doptics[key_cam]['paths']
    else:
        dap = dap0

    if timing:
        t11 = dtm.datetime.now()     # DB
        dt11 += (t11-t00).total_seconds()

    # -----------------
    # initialize lists
    # -----------------

    (
        # common
        lpcross, lphor, lsang_cross, lindr_cross, lindz_cross,
        # keep3d
        lindr_3d, lindz_3d, lphi_3d, lsang_3d,
        # return_vector
        lang_tor_cross, lang_pol_cross, lvectx, lvecty, lvectz
    ) = _initialize_lists(
        return_vector=return_vector,
        keep3d=keep3d,
    )

    # ----------------
    # loop on pixels
    # ----------------

    shape_cam = coll.dobj['camera'][key_cam]['dgeom']['shape']
    npix = int(np.prod(shape_cam))
    linds = [range(ss) for ss in shape_cam]
    for ii, ind in enumerate(itt.product(*linds)):

        debugi = debug if isinstance(debug, bool) else debug(ind)

        # -----------------
        # slices

        sli_poly = tuple([slice(None)] + list(ind))
        sli_poly0 = tuple([0] + list(ind))

        # -----------------
        # get volume limits

        if timing:
            t000 = dtm.datetime.now()     # DB

        # get points
        if user_limits is None or user_limits.get('pcross_user') is None:

            if np.isnan(pcross0[sli_poly0]):
                pts_cross = np.zeros((0,), dtype=float)
                lpcross.append((None, None))
                lphor.append((None, None))
                lsang_cross.append(pts_cross)
                lindr_cross.append(pts_cross)
                lindz_cross.append(pts_cross)
                if return_vector is True:
                    lang_pol_cross.append(pts_cross)
                    lang_tor_cross.append(pts_cross)
                continue

            xx, yy, zz, dind, ir, iz, iphi, dV = _vos_points(
                # polygons
                pcross0=pcross0[sli_poly],
                pcross1=pcross1[sli_poly],
                phor0=phor0[sli_poly],
                phor1=phor1[sli_poly],
                margin_poly=margin_poly,
                dphi=dphi[sli_poly],
                # sampling
                dsamp=dsamp,
                x0f=x0f,
                x1f=x1f,
                x0u=x0u,
                x1u=x1u,
                res=res_phi,
                dx0=dx0,
                dx1=dx1,
                # shape
                sh=sh,
                # debug
                debug=debugi,
                ii=ii,
                ind=ind,
            )

        if xx is None:
            pts_cross = np.zeros((0,), dtype=float)
            lpcross.append((None, None))
            lphor.append((None, None))
            lsang_cross.append(pts_cross)
            lindr_cross.append(pts_cross)
            lindz_cross.append(pts_cross)
            if return_vector is True:
                lang_pol_cross.append(pts_cross)
                lang_tor_cross.append(pts_cross)
            continue

        # --------------
        # re-initialize

        bool_cross[...] = False
        npts_tot = xx.size
        npts_cross = np.sum([v0['iz'].size for v0 in dind.values()])

        dsang_hor = {}
        sang_cross = np.zeros((npts_cross,), dtype=float)
        indr_cross = np.zeros((npts_cross,), dtype=int)
        indz_cross = np.zeros((npts_cross,), dtype=int)
        if return_vector is True:
            ang_pol_cross = np.zeros((npts_cross,), dtype=float)
            ang_tor_cross = np.zeros((npts_cross,), dtype=float)

        if verb is True:
            msg = (
                f"\tcam '{key_cam}' pixel {ii+1} / {npix}"
                f"\tnpts in cross_section = {npts_cross}"
                f"\t({npts_tot} total)"
            )
            end = '\n 'if ii == npix - 1 else '\r'
            print(msg, end=end, flush=True)

        # --------------------------------
        # get formatted detector geometry

        # get detector / aperture
        deti = _get_deti(
            coll=coll,
            cxi=cx[ind],
            cyi=cy[ind],
            czi=cz[ind],
            dvect=dvect,
            par=par,
            out0=out0,
            out1=out1,
            ind=ind,
        )

        if timing:
            t111 = dtm.datetime.now()     # DB
            dt111 += (t111-t000).total_seconds()

        # --------------------------------
        # get pixel-specific apertures if not pinhole

        if pinhole is False:
            sli_path = tuple(list(ind) + [slice(None)])
            iop = np.nonzero(paths[sli_path])[0]
            dap = {optics[ii]: dap0[optics[ii]] for ii in iop}

        # -------------------
        # compute solid angle

        # compute
        out = _comp_solidangles.calc_solidangle_apertures(
            # observation points
            pts_x=xx,
            pts_y=yy,
            pts_z=zz,
            # polygons
            apertures=dap,
            detectors=deti,
            # possible obstacles
            config=config,
            # parameters
            summed=False,
            visibility=visibility,
            return_vector=return_vector,
            return_flat_pts=None,
            return_flat_det=None,
            timing=timing,
        )

        if isinstance(out, tuple):
            out, vectx, vecty, vectz = out
        else:
            vectx, vecty, vectz = None, None, None

        # ------------
        # get indices

        if timing:
            t0 = dtm.datetime.now()     # DB
            out, dt1, dt2, dt3 = out

        # update cross-section
        ipt = 0
        for i0, v0 in dind.items():
            for i1 in v0['iz']:
                ind1 = dind[i0]['indrz'] & (iz == i1)
                totii = np.sum(out[0, ind1]) * v0['dV']
                sang_cross[ipt] = totii
                indr_cross[ipt] = i0
                indz_cross[ipt] = i1
                bool_cross[i0 + 1, i1 + 1] = sang_cross[ipt] > 0.
                if vectx is not None:
                    tor = np.arctan2(yy[ind1], xx[ind1])
                    ang_pol = np.arctan2(
                        vectz[0, ind1],
                        vectx[0, ind1]*np.cos(tor) + vecty[0, ind1]*np.sin(tor),
                    )
                    ang_tor = np.arccos(
                        vectx[0, ind1] * (-np.sin(tor))
                        + vecty[0, ind1] * np.cos(tor)
                    )
                    ang_pol_cross[ipt] = np.sum(out[0, ind1] * ang_pol) / totii
                    ang_tor_cross[ipt] = np.sum(out[0, ind1] * ang_tor) / totii
                ipt += 1

        # update horizontal
        for i0, v0 in dind.items():
            dsang_hor[i0] = np.zeros((v0['phi'].size,))
            for i1 in range(v0['phi'].size):
                ind1 = dind[i0]['indrz'] & (iphi == i1)
                dsang_hor[i0][i1] = np.sum(out[0, ind1]) * v0['dV']

        # update 3d
        if keep3d is True and np.any(bool_cross):
            indsa = out[0, :] > 0.

            indr_3d = ir[indsa]
            indz_3d = iz[indsa]
            phi_3d = np.arctan2(yy[indsa], xx[indsa])
            sang_3d = out[0, indsa] * dV[indsa]

            if vectx is not None:
                vx = vectx[0, indsa]
                vy = vecty[0, indsa]
                vz = vectz[0, indsa]

        # ----- DEBUG --------
        if debugi:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            fig.suptitle(f"pixel ind = {ind}", size=14, fontweight='bold')
            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
            ax.set_xlabel("phi (deg)", size=12)
            ax.set_ylabel("solid angle (sr)", size=12)
            # ipos = out[0, :] > 0
            # ax.scatter(
            #     xx[ipos], yy[ipos],
            #     c=out[0, ipos], s=6, marker='o', vmin=0,
            # )
            # ax.plot(xx[~ipos], yy[~ipos], c='r', marker='x')
            ax.scatter(
                np.arctan2(yy, xx) * 180/np.pi,
                out[0, :],
                c=np.hypot(xx, yy),
                s=6,
                marker='.',
            )
            raise Exception()
        # ----- END DEBUG ----

        # timing
        if timing:
            dt4444 += (dtm.datetime.now() - t0).total_seconds()
            dt1111 += dt1
            dt2222 += dt2
            dt3333 += dt3
            t222 = dtm.datetime.now()     # DB
            dt222 += (t222-t111).total_seconds()

        # -----------------------
        # get pcross and simplify

        if np.any(bool_cross):

            # pcross
            pc0, pc1 = _utilities._get_polygons(
                bool_cross=bool_cross,
                x0=x0l,
                x1=x1l,
                res=np.min(np.atleast_1d(res_RZ)),
            )

            ph0, ph1 = _get_phor2(
                xx=xx,
                yy=yy,
                out=out[0, :],
                res=min(res_RZ[0], res_phi),
            )

        else:
            pc0, pc1 = None, None
            ph0, ph1 = None, None

        # -----------
        # replace

        lpcross.append((pc0, pc1))
        lphor.append((ph0, ph1))
        lsang_cross.append(sang_cross)
        lindr_cross.append(indr_cross)
        lindz_cross.append(indz_cross)
        if lang_pol_cross is not None:
            lang_pol_cross.append(ang_pol_cross)
            lang_tor_cross.append(ang_tor_cross)

        if keep3d is True:
            lsang_3d.append(sang_3d)
            lindr_3d.append(indr_3d)
            lindz_3d.append(indz_3d)
            lphi_3d.append(phi_3d)

            if lvectx is not None:
                lvectx.append(vx)
                lvecty.append(vy)
                lvectz.append(vz)

        if timing:
            t333 = dtm.datetime.now()     # DB
            dt333 += (t333-t222).total_seconds()

    # ----------------------------
    # harmonize and reshape pcross
    # ----------------------------

    if timing:
        t22 = dtm.datetime.now()     # DB

    pcross0, pcross1 = _harmonize_reshape_pcross(
        lpcross=lpcross,
        shape=shape_cam,
    )

    phor0, phor1 = _harmonize_reshape_pcross(
        lpcross=lphor,
        shape=shape_cam,
    )

    # --------------------------------------
    # harmonize and reshape sang, indr, indz
    # --------------------------------------

    # cross
    dout = _harmonize_reshape_others(
        # cross
        lsang_cross=lsang_cross,
        lindr_cross=lindr_cross,
        lindz_cross=lindz_cross,
        # params
        npix=npix,
        is2d=is2d,
        shape=shape_cam,
    )

    # extract
    lk = ['lsang_cross', 'lindr_cross', 'lindz_cross']
    sang_cross, indr_cross, indz_cross = [dout.get(k0) for k0 in lk]


    if lang_pol_cross is not None:
        dout = _harmonize_reshape_others(
            lang_pol_cross=lang_pol_cross,
            lang_tor_cross=lang_tor_cross,
            # params
            npix=npix,
            is2d=is2d,
            shape=shape_cam,
        )
        lk = ['lang_pol_cross', 'lang_tor_cross']
        ang_pol_cross, ang_tor_cross = [dout.get(k0) for k0 in lk]

    # -------
    # 3d

    dout = _harmonize_reshape_others(
        # 3d
        lsang_3d=lsang_3d,
        lindr_3d=lindr_3d,
        lindz_3d=lindz_3d,
        lphi_3d=lphi_3d,
        # vect
        lvectx=lvectx,
        lvecty=lvecty,
        lvectz=lvectz,
        # params
        npix=npix,
        is2d=is2d,
        shape=shape_cam,
    )

    # extract
    lk = [
        'lsang_3d', 'lindr_3d', 'lindz_3d', 'lphi_3d',
        'lvectx', 'lvecty', 'lvectz',
    ]

    (
     sang_3d, indr_3d, indz_3d, phi_3d,
     vectx, vecty, vectz,
     ) = [dout.get(k0) for k0 in lk]

    # --------------
    # prepare output
    # --------------

    knpts_cross = f'{key_cam}_vos_npts_cross'
    kir_cross = f'{key_cam}_vos_ir_cross'
    kiz_cross = f'{key_cam}_vos_iz_cross'
    ksa_cross = f'{key_cam}_vos_sa_cross'
    ref_cam = coll.dobj['camera'][key_cam]['dgeom']['ref']
    ref_cross = tuple(list(ref_cam) + [knpts_cross])

    if lang_pol_cross is not None:
        kap_cross = f'{key_cam}_vos_ang_pol_cross'
        kat_cross = f'{key_cam}_vos_ang_tor_cross'

    if keep3d:
        knpts_3d = f'{key_cam}_vos_npts_3d'
        kir_3d = f'{key_cam}_vos_ir_3d'
        kiz_3d = f'{key_cam}_vos_iz_3d'
        kphi_3d = f'{key_cam}_vos_phi_3d'
        ksa_3d = f'{key_cam}_vos_sa_3d'
        ref_3d = tuple(list(ref_cam) + [knpts_3d])

        if lvectx is not None:
            kvectx = f'{key_cam}_vos_vx'
            kvecty = f'{key_cam}_vos_vy'
            kvectz = f'{key_cam}_vos_vz'

    # ----------------
    # format output
    # ----------------

    # dref
    dref = {
        'npts_cross': {
            'key': knpts_cross,
            'size': indr_cross.shape[-1],
        },
    }

    if keep3d is True:
        dref['npts_3d'] = {
            'key': knpts_3d,
            'size': indr_3d.shape[1],
        }

    # ddata - polygons
    dout = {
        'pcross0': {
            'data': pcross0,
            'units': 'm',
            'dim': 'distance',
        },
        'pcross1': {
            'data': pcross1,
            'units': 'm',
            'dim': 'distance',
        },
        'phor0': {
            'data': phor0,
            'units': 'm',
            'dim': 'distance',
        },
        'phor1': {
            'data': phor1,
            'units': 'm',
            'dim': 'distance',
        },
    }

    # ddata - cross indices
    dout.update({
        'indr_cross': {
            'key': kir_cross,
            'data': indr_cross,
            'ref': ref_cross,
            'units': '',
            'dim': 'index',
        },
        'indz_cross': {
            'key': kiz_cross,
            'data': indz_cross,
            'ref': ref_cross,
            'units': '',
            'dim': 'index',
        },
        'sang_cross': {
            'key': ksa_cross,
            'data': sang_cross,
            'ref': ref_cross,
            'units': 'sr.m3',
            'dim': 'sang',
        },
    })

    # return_vect
    if lang_pol_cross is not None:
        dout.update({
            'ang_pol_cross': {
                'key': kap_cross,
                'data': ang_pol_cross,
                'ref': ref_cross,
                'units': 'rad',
                'dim': 'angle',
            },
            'ang_tor_cross': {
                'key': kat_cross,
                'data': ang_tor_cross,
                'ref': ref_cross,
                'units': 'rad',
                'dim': 'angle',
            },
        })

    # keep3d
    if keep3d is True:
        dout.update({
            'indr_3d': {
                'key': kir_3d,
                'data': indr_3d,
                'ref': ref_3d,
                'units': '',
                'dim': 'index',
            },
            'indz_3d': {
                'key': kiz_3d,
                'data': indz_3d,
                'ref': ref_3d,
                'units': '',
                'dim': 'index',
            },
            'phi_3d': {
                'key': kphi_3d,
                'data': phi_3d,
                'ref': ref_3d,
                'units': '',
                'dim': 'index',
            },
            'sang_3d': {
                'key': ksa_3d,
                'data': sang_3d,
                'ref': ref_3d,
                'units': 'sr.m3',
                'dim': 'sang',
            },
        })

        if lvectx is not None:
            dout.update({
                'vectx_3d': {
                    'key': kvectx,
                    'data': vectx,
                    'ref': ref_3d,
                    'units': '',
                },
                'vecty_3d': {
                    'key': kvecty,
                    'data': vecty,
                    'ref': ref_3d,
                    'units': '',
                },
                'vectz_3d': {
                    'key': kvectz,
                    'data': vectz,
                    'ref': ref_3d,
                    'units': '',
                },
            })

    if timing:
        t33 = dtm.datetime.now()
        dt22 += (t33 - t22).total_seconds()

    return (
        dout, dref,
        dt11, dt22,
        dt111, dt222, dt333,
        dt1111, dt2222, dt3333, dt4444,
    )


# ###########################################################
# ###########################################################
#               Iniialize lists
# ###########################################################


def _initialize_lists(
    return_vector=None,
    keep3d=None,
):

    lpcross = []
    lphor = []
    lsang_cross = []
    lindr_cross = []
    lindz_cross = []
    if keep3d is True:
        lindr_3d = []
        lindz_3d = []
        lphi_3d = []
        lsang_3d = []
    else:
        lindr_3d = None
        lindz_3d = None
        lphi_3d = None
        lsang_3d = None

    if return_vector is True:
        lang_tor_cross = []
        lang_pol_cross = []
        if keep3d is True:
            lvectx = []
            lvecty = []
            lvectz = []
        else:
            lvectx = None
            lvecty = None
            lvectz = None
    else:
        lang_tor_cross = None
        lang_pol_cross = None
        lvectx = None
        lvecty = None
        lvectz = None

    return (
        lpcross, lphor, lsang_cross, lindr_cross, lindz_cross,
        lindr_3d, lindz_3d, lphi_3d, lsang_3d,
        lang_tor_cross, lang_pol_cross, lvectx, lvecty, lvectz,
    )


# ###########################################################
# ###########################################################
#               get points
# ###########################################################


def _vos_points(
    # polygons
    pcross0=None,
    pcross1=None,
    phor0=None,
    phor1=None,
    margin_poly=None,
    dphi=None,
    # sampling
    dsamp=None,
    x0f=None,
    x1f=None,
    x0u=None,
    x1u=None,
    res=None,
    dx0=None,
    dx1=None,
    # shape
    sh=None,
    # debug
    debug=None,
    ii=None,
    ind=None,
):

    # ---------------------
    # get polygons - cross
    # ---------------------

    # ---------------------
    # get cross-section polygon with margin

    pc0, pc1 = _utilities._get_poly_margin(
        # polygon
        p0=pcross0,
        p1=pcross1,
        # margin
        margin=margin_poly,
    )

    pcross = Path(np.array([pc0, pc1]).T)

    # ------------
    # get indices
    # ------------

    # indices
    index = (
        dsamp['ind']['data']
        & pcross.contains_points(np.array([x0f, x1f]).T).reshape(sh)
    )

    # R and Z indices
    ir, iz = index.nonzero()
    iru = np.unique(ir)

    # ---------------------------
    # get polygons - cross dphi_r
    # ---------------------------

    # get cross-section polygon with margin
    ph0, ph1 = _utilities._get_poly_margin(
        # polygon
        p0=phor0,
        p1=phor1,
        # margin
        margin=margin_poly,
    )

    # ------------
    # get dphi_r
    # ------------

    # phi_r
    dphi_r = _utilities._get_dphi_from_R_phor(
        R=x0u[iru],
        phor0=ph0,
        phor1=ph1,
        phimin=dphi[0],
        phimax=dphi[1],
        res=res,
    )

    # get nphi
    iok = np.all(np.isfinite(dphi_r), axis=0)
    dphi_r = dphi_r[:, iok]
    iru = iru[iok]

    # ------------
    # safety check
    # ------------

    if iru.size == 0:
        return None, None, None, None, None, None, None, None

    # ------------
    # go on
    # ------------

    nphi_r = (
        np.ceil(x0u[iru]*(dphi_r[1, :] - dphi_r[0, :]) / res).astype(int)
        + 1
    )
    ddphi_r = np.diff(dphi_r, axis=0)[0, :] / (nphi_r - 1)

    # ------------
    # get indices
    # ------------

    # get indices
    lind = [ir == i0 for i0 in iru]
    ln = [i0.sum() for i0 in lind]

    indrz = np.concatenate([
        np.tile(i0.nonzero()[0], nphi_r[ii]) for ii, i0 in enumerate(lind)
    ])

    # get phi
    lphi = [
        np.linspace(dphi_r[0, ii], dphi_r[1, ii], nn)
        for ii, nn in enumerate(nphi_r)
    ]
    phi = np.concatenate(tuple([
        np.repeat(phii, ln[ii]) for ii, phii in enumerate(lphi)
    ]))
    iphi = np.concatenate(tuple([
        np.repeat(np.arange(0, nn), ln[ii]) for ii, nn in enumerate(nphi_r)
    ]))

    dV = np.concatenate(tuple([
        np.repeat(dx0 * dx1 * x0u[i0] * ddphi_r[ii], ln[ii] * nphi_r[ii])
        for ii, i0 in enumerate(iru)
    ]))

    # -------------
    # derive coords
    # -------------

    # coordinates
    rr = x0u[ir[indrz]]
    xx = rr * np.cos(phi)
    yy = rr * np.sin(phi)
    zz = x1u[iz[indrz]]

    # ----------------
    # get indices dict
    # ----------------

    dind = {
        i0: {
            'dV': dx0 * dx1 * x0u[i0] * ddphi_r[ii],
            'iz': np.unique(iz[lind[ii]]),
            'indrz': ir[indrz] == i0,
            'phi': lphi[ii],
        }
        for ii, i0 in enumerate(iru)
    }

    # ----------------
    # debug
    # ----------------

    if debug is True:

        fig = plt.figure(figsize=(14, 8))
        fig.suptitle(f"pixel ind = {ind}", size=14, fontweight='bold')

        ax0 = fig.add_subplot(1, 2, 1, aspect='equal')
        ax0.set_xlabel("R (m)", size=12, fontweight='bold')
        ax0.set_ylabel("Z (m)", size=12, fontweight='bold')

        ax1 = fig.add_subplot(1, 2, 2, aspect='equal')
        ax1.set_xlabel("X (m)", size=12, fontweight='bold')
        ax1.set_ylabel("Y (m)", size=12, fontweight='bold')

        ax0.fill(pc0, pc1, fc=(0.5, 0.5, 0.5, 0.5))
        ax1.fill(ph0, ph1, fc=(0.5, 0.5, 0.5, 0.5))

        ax0.plot(np.hypot(xx, yy), zz, '.')
        ax1.plot(xx, yy, '.')

    return xx, yy, zz, dind, ir[indrz], iz[indrz], iphi, dV


# ###########################################################
# ###########################################################
#               PHOR
# ###########################################################


def _get_phor2(xx=None, yy=None, out=None, res=None, debug=False):

    # boundaries
    xmin, xmax = xx.min(), xx.max()
    ymin, ymax = yy.min(), yy.max()

    # grid
    nx = int(np.ceil((xmax - xmin) / (2 * res)))
    ny = int(np.ceil((ymax - ymin) / (2 * res)))

    xb = np.linspace(xmin - 4*res, xmax + 4*res, nx+4)
    yb = np.linspace(ymin - 4*res, ymax + 4*res, ny+4)

    binned = scpstats.binned_statistic_2d(
        xx,
        yy,
        out,
        bins=(xb, yb),
        statistic='mean',
    ).statistic

    # ------- DEBUG -------
    if debug:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.scatter(xx, yy, c=out, s=6, marker='.')

        fig = plt.figure()
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.imshow(
            binned.T,
            extent=(xb[0], xb[-1], yb[0], yb[-1]),
            aspect='equal',
            interpolation='nearest',
            origin='lower',
        )
    # -----------------------

    # --------------------------
    # get phor in (r, phi) space

    phx, phy = _utilities._get_polygons(
        bool_cross=binned > 0,
        x0=np.repeat(0.5*(xb[1:] + xb[:-1])[:, None], ny+3, axis=1),
        x1=np.repeat(0.5*(yb[1:] + yb[:-1])[None, :], nx+3, axis=0),
        res=2*res,
    )

    return phx, phy


# DEPRECATED ?
# def _get_phor(dind=None, dsang_hor=None, x0=None, res=None):

#     # ------------
#     # get phi map

#     dphi = np.min([
#         (v0['phi'][1] - v0['phi'][0]) for v0 in dind.values()
#         if v0['phi'].size >= 2
#     ])

#     phi_min = np.min([np.min(v0['phi']) for v0 in dind.values()])
#     phi_max = np.max([np.max(v0['phi']) for v0 in dind.values()])

#     nphi = int(np.ceil((phi_max - phi_min) / dphi))
#     phi = np.linspace(phi_min - dphi, phi_max + dphi, nphi + 2)

#     # --------------
#     # get sang map

#     nr = x0.size
#     bool_hor = np.zeros((nr, nphi + 2), dtype=float)

#     for ii, (i0, v0) in enumerate(dind.items()):

#         bool_hor[i0 + 1, :] = scpinterp.UnivariateSpline(
#             v0['phi'],
#             dsang_hor[i0],
#             w=None,
#             bbox=[None, None],
#             k=1,
#             s=None,
#             ext=0,
#             check_finite=False,
#         )(phi) > 0
#     bool_hor[:, 0] = 0.
#     bool_hor[:, -1] = 0.

#     # ----------------
#     # convert to x, y

#     rf = np.repeat(x0[:, None], nphi+2, axis=1)
#     phif = np.repeat(phi[None, :], nr, axis=0)

#     xf = rf * np.cos(phif)
#     yf = rf * np.sin(phif)

#     xmin, xmax = xf.min(), xf.max()
#     ymin, ymax = yf.min(), yf.max()

#     res = min(res, dphi*x0[0])
#     nx = int(np.ceil((xmax - xmin) / res))
#     ny = int(np.ceil((ymax - ymin) / res))

#     xx = np.linspace(np.min(xf), np.max(xf), nx)
#     yy = np.linspace(np.min(xf), np.max(xf), ny)
#     rr = np.hypot(xx[:, None], yy[None, :])
#     pp = np.arctan2(yy[None, :], xx[:, None])

#     bool_xy = np.zeros((nx, ny), dtype=float)

#     iok = (rr > x0[0]) & (rr < x0[-1])
#     bool_xy[iok] = scpinterp.RectBivariateSpline(
#         x0,
#         phi,
#         bool_hor,
#         kx=1,
#         ky=1,
#         s=0,
#     )(rr[iok], pp[iok], grid=False)

#     # --------------------------
#     # get phor in (r, phi) space

#     phx, phy = _utilities._get_polygons(
#         bool_cross=bool_xy,
#         x0=xx,
#         x1=yy,
#         res=res,
#     )

#     return phx, phy


# ###########################################################
# ###########################################################
#               Pixel
# ###########################################################


# def _vos_pixel(
#     x0=None,
#     x1=None,
#     ind=None,
#     npts=None,
#     dphi=None,
#     deti=None,
#     lap=None,
#     res=None,
#     config=None,
#     visibility=None,
#     # output
#     key_cam=None,
#     sli=None,
#     ii=None,
#     bool_cross=None,
#     sang=None,
#     indr=None,
#     indz=None,
#     # timing
#     timing=None,
#     dt1111=None,
#     dt2222=None,
#     dt3333=None,
#     dt4444=None,
# ):


#     out = _comp_solidangles.calc_solidangle_apertures(
#         # observation points
#         pts_x=xx,
#         pts_y=yy,
#         pts_z=zz,
#         # polygons
#         apertures=lap,
#         detectors=deti,
#         # possible obstacles
#         config=config,
#         # parameters
#         summed=False,
#         visibility=visibility,
#         return_vector=False,
#         return_flat_pts=None,
#         return_flat_det=None,
#         timing=timing,
#     )

#     # ------------
#     # get indices

#     if timing:
#         t0 = dtm.datetime.now()     # DB
#         out, dt1, dt2, dt3 = out

#     ipt = 0
#     for ii, i0 in enumerate(iru):
#         ind0 = irf == i0
#         for i1 in izru[ii]:
#             ind = ind0 & (izf == i1)
#             bool_cross[i0 + 1, i1 + 1] = np.any(out[0, ind] > 0.)
#             sang[ipt] = np.sum(out[0, ind])
#             indr[ipt] = i0
#             indz[ipt] = i1
#             ipt += 1
#     assert ipt == npts

#     # timing
#     if timing:
#         dt4444 += (dtm.datetime.now() - t0).total_seconds()
#         dt1111 += dt1
#         dt2222 += dt2
#         dt3333 += dt3

#         return dt1111, dt2222, dt3333, dt4444
#     else:
#         return


# ###########################################################
# ###########################################################
#               Detector
# ###########################################################


def _get_deti(
    coll=None,
    cxi=None,
    cyi=None,
    czi=None,
    dvect=None,
    par=None,
    out0=None,
    out1=None,
    ind=None,
):

    # ------------
    # detector

    det = {
        'cents_x': cxi,
        'cents_y': cyi,
        'cents_z': czi,
        'outline_x0': out0,
        'outline_x1': out1,
        'nin_x': dvect['nin_x'] if par else dvect['nin_x'][ind],
        'nin_y': dvect['nin_y'] if par else dvect['nin_y'][ind],
        'nin_z': dvect['nin_z'] if par else dvect['nin_z'][ind],
        'e0_x': dvect['e0_x'] if par else dvect['e0_x'][ind],
        'e0_y': dvect['e0_y'] if par else dvect['e0_y'][ind],
        'e0_z': dvect['e0_z'] if par else dvect['e0_z'][ind],
        'e1_x': dvect['e1_x'] if par else dvect['e1_x'][ind],
        'e1_y': dvect['e1_y'] if par else dvect['e1_y'][ind],
        'e1_z': dvect['e1_z'] if par else dvect['e1_z'][ind],
    }

    return det


# ###########################################################
# ###########################################################
#               Harmonize and reshape pcross
# ###########################################################


def _harmonize_reshape_pcross(
    lpcross=None,
    shape=None,
):

    # ------------------
    # get max nb of pts
    # ------------------

    # list if nb of pts
    ln = [pp[0].size if pp[0] is not None else 0 for pp in lpcross]
    nmax = np.max(ln)

    # initialize
    sh2 = tuple([nmax] + list(shape))
    pcross0 = np.full(sh2, np.nan)
    pcross1 = np.full(sh2, np.nan)

    # ----------------
    # loop on pixels
    # ---------------

    linds = [range(ss) for ss in shape]
    for ii, ind in enumerate(itt.product(*linds)):

        nn = ln[ii]
        if nn == 0:
            continue

        sli = tuple([slice(None)] + list(ind))

        if nmax > nn:
            ind = np.r_[0, np.linspace(0.1, 0.9, nmax - nn), np.arange(1, nn)]
            pcross0[sli] = scpinterp.interp1d(
                range(0, nn),
                lpcross[ii][0],
                kind='linear',
            )(ind)

            pcross1[sli] = scpinterp.interp1d(
                range(0, nn),
                lpcross[ii][1],
                kind='linear',
            )(ind)

        else:
            pcross0[sli] = lpcross[ii][0]
            pcross1[sli] = lpcross[ii][1]

    return pcross0, pcross1


# ###########################################################
# ###########################################################
#               Harmonize and reshape others
# ###########################################################


def _harmonize_reshape_others(
    npix=None,
    is2d=None,
    shape=None,
    **kwdargs,
):

    # -----------------------------
    # check all input args have the same list of sizes
    # -----------------------------

    dnpts = {
        k0: [v1.size for v1 in v0] for k0, v0 in kwdargs.items()
        if v0 is not None
    }
    lkey = list(dnpts.keys())

    if not all([np.allclose(v0, dnpts[lkey[0]]) for v0 in dnpts.values()]):
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dnpts.items()]
        msg = (
            "All input args must have the same list of sizes!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # ------
    # trival

    if len(lkey) == 0:
        return {}

    # list of sizes and max size
    lnpts = dnpts[lkey[0]]
    nmax = np.max(lnpts)
    shape_max = tuple(list(shape) + [nmax])

    # ------------------
    # prepare
    # ------------------

    dout = {}
    for k0 in lkey:

        # initialize
        if 'ind' in k0:
            dout[k0] = -np.ones(shape_max, dtype=int)
        else:
            dout[k0] = np.full(shape_max, np.nan)

        # fill
        linds = [range(ss) for ss in shape]
        for ii, ind in enumerate(itt.product(*linds)):
            sli = tuple(list(ind) + [np.arange(lnpts[ii])])
            dout[k0][sli] = kwdargs[k0][ii]

    return dout


"""
Computing the integrated solid angle of a particle
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
This tutorial show how to crete a configuration and computing the
integrated solid angle subtended by a particle's trajectory along phi
with certain discretization, and plotting it.

:math:`\int_\Phi \Omega R d \Phi`

Useful for reconstructing emissivity.
"""
###############################################################################
# We start by loading ITER's configuration (built-in in `tofu`)

import matplotlib.pyplot as plt
import numpy as np
import tofu as tf


config = tf.load_config("ITER")

###############################################################################
# We define the particles properties and trajectory.
# Let's suppose we have the data for three points of the trajectory,
# the particle moves along X from point (5, 0, 0) in cartesian coordinates,
# to (6, 0, 0) and finally to (7, 0, 0). At the end point the particle radius
# seems to be a bit bigger (:math:`2 \mu m` instead of :math:`1 \mu m`)

part_rad = np.r_[.0001, .0001, .0002]*2

part_traj = np.array([[5.0, 0.0, 0.0],
                      [6.0, 0.0, 0.0],
                      [7.0, 0.0, 0.0]], order="F").T

###############################################################################
# Let's set some parameters for the discretization for computing the integral:
# resolutions in (R, Z, Phi) directions and the values to only compute the
# integral on the core of the plasma.

r_step = z_step = phi_step = 0.02  # 1 cm resolution along all directions
Rminmax = np.r_[4.0, 8.0]
Zminmax = np.r_[-5., 5.0]

###############################################################################
# Let's compute the integrated solid angle: the function returns the points
# in (R,Z) of the discretization, the integrated solid angle on those points,
# the indices to reconstruct the discretization in all domain and a the volume
# unit :math:`dR * dZ`.

pts, sa_map, ind, vol = tf.geom._GG.compute_solid_angle_map(part_traj,
                                                            part_rad,
                                                            r_step,
                                                            z_step,
                                                            phi_step,
                                                            Rminmax,
                                                            Zminmax)

###############################################################################
# Now we can plot the results the first point on the trajectory
fig1, ax = plt.subplots()

ax.scatter(pts[0, :], pts[1, :],    # R and Z coordinates
           marker="s",              # each point is a squared pixel
           edgecolors=None,         # no boundary for smooth plot
           s=10,                    # size of pixel
           c=sa_map[:, 0].flatten(),  # pixel color is value of int solid angle
           )
ax.set_aspect("equal")
plt.show()

###############################################################################
# or the three points in the trajectory
fig2, list_axes = plt.subplots(ncols=3, sharey=True)

# Now we can plot the results for all points on the trajectory
for (ind, ax) in enumerate(list_axes):
    ax.scatter(pts[0, :], pts[1, :],
               marker="s",
               edgecolors=None,
               s=10,
               c=sa_map[:, ind].flatten(),  # we change particle number
               )
    ax.set_aspect("equal")


plt.show()

###############################################################################
# Now let's see the 1D profile of all particles for z = 0.
izero = np.abs(pts[1, :]) < z_step

fig3, list_axes = plt.subplots(ncols=3, sharey=False)

for (ind, ax) in enumerate(list_axes):
    ax.plot(pts[0, izero], sa_map[izero, ind])

plt.show()



# -*- coding: utf-8 -*-


# Common
import numpy as np
import scipy.interpolate as scpinterp
import scipy.integrate as scpinteg
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import datastock as ds


# ###############################################################
# ###############################################################
#                       Main
# ###############################################################


def plot_rocking_curve(
    coll=None,
    key=None,
    # option
    T=None,
    # plotting
    dax=None,
    color=None,
    plot_FW=None,
):

    # -------------
    # check inputs

    # key
    lok = [
        k0 for k0, v0 in coll.dobj.get('crystal', {}).items()
        if v0.get('dmat', {}).get('drock') is not None
    ]

    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # drock
    drock = coll.dobj['crystal'][key]['dmat']['drock']
    is2d = drock.get('Tref') is not None

    # color
    if color is None:
        color = 'k'

    # -----------
    # plot

    # is2d
    if is2d:

        if T is False:
            T = drock['Tref']
            bragg = None
            is2d = False

        elif T is not None:
            T = ds._generic_check._check_var(
                T, 'T',
                types=(int, float),
                sign='>0',
            )

            T0 = coll.ddata[drock['T']]['data']
            braggT = coll.ddata[drock['braggT']]['data']
            indT = np.argmin(np.abs(T0 - T))
            bragg = braggT[indT]
            is2d = False

        else:
            return _plot_rc_2d(
                key=key,
                coll=coll,
                drock=drock,
                # plotting
                dax=dax,
                color=color,
            )

    else:
        T = None

    # not is2d
    if is2d is False:

        if bragg is None:
            bragg = coll.get_crystal_bragglamb(
                key=key,
                rocking_curve=False,
            )[0][0]

        return _plot_rc_1d(
            key=key,
            coll=coll,
            drock=drock,
            bragg=bragg,
            T=None,
            # plotting
            dax=dax,
            color=color,
            plot_FW=plot_FW,
        )


# ###############################################################
# ###############################################################
#                       Plot 1d
# ###############################################################


def _plot_rc_1d(
    key=None,
    coll=None,
    drock=None,
    bragg=None,
    T=None,
    # plotting
    dax=None,
    color=None,
    plot_FW=None,
):

    # -------
    # check

    if plot_FW is None:
        plot_FW = True

    if issubclass(dax.__class__, plt.Axes):
        dax = {'rc': {'handle': dax, 'type': 'matrix'}}

    # ----------
    # prepare

    lamb = coll.dobj['crystal'][key]['dmat']['target']['lamb']
    angle_rel = coll.ddata[drock['angle_rel']]['data']
    power_ratio = coll.ddata[drock['power_ratio']]['data']
    delta_bragg = drock['delta_bragg']

    integ = drock['integ_reflect']
    pmax = np.max(power_ratio)
    FW = integ / pmax
    imax = np.argmax(power_ratio)
    xFW = angle_rel[imax] + FW/2.*np.r_[-1, 1]

    xFW05 = angle_rel[power_ratio > 0.05*pmax]

    if T is None:
        lab = f'{key} (R = {integ} rad)'
    else:
        lab = f'{key} (R = {integ} rad at T = {T} C)'

    # ----------
    # figure

    if dax is None:
        fig = plt.figure(figsize=(12, 12))
        fig.suptitle('Rocking curve', size=14, fontweight='bold')

        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_xlabel('incidence angle relative (rad)')
        ax.set_ylabel('power ratio')
        ax.set_ylim(0, 1)


        dax = {'main': {'handle': ax, 'type': 'matrix'}}

    # ----------
    # plot

    axtype = 'matrix'
    lax = [k0 for k0, v0 in dax.items() if v0['type'] == axtype]
    if len(lax) > 0:
        kax = lax[0]
        ax = dax[kax]['handle']

        ax.plot(
            angle_rel,
            power_ratio,
            ls='-',
            marker='.',
            lw=1.,
            c=color,
            label=lab,
        )

        # bragg angle
        ax.axvline(0, c=color, ls='--', lw=2.)
        ax.axvline(delta_bragg, c=color, ls='--', lw=1)

        ax.text(
            0,
            1,
            f'bragg({lamb*1e10:.3f} A)\n= {bragg:.6f} rad',
            horizontalalignment='center',
            verticalalignment='bottom',
            color=color,
        )

        if plot_FW:
            fc = tuple(list(mcolors.to_rgb(color)) + [0.3])

            ax.fill_between(
                xFW,
                np.zeros((xFW.size,)),
                pmax*np.ones((xFW.size,)),
                fc=fc,
            )

            ax.fill_between(
                xFW05,
                np.zeros((xFW05.size,)),
                0.05*pmax*np.ones((xFW05.size,)),
                fc=fc,
            )
    return ax


# ###############################################################
# ###############################################################
#                       Plot 2d
# ###############################################################


def _plot_rc_2d(
    key=None,
    coll=None,
    drock=None,
    # plotting
    dax=None,
    color=None,
):

    # ----------
    # prepare

    angle_rel = coll.ddata[drock['angle_rel']]['data']
    power_ratio = coll.ddata[drock['power_ratio']]['data']
    na = angle_rel.size

    integ = drock['integ_reflect']

    T = coll.ddata[drock['T']]['data']
    braggT = coll.ddata[drock['braggT']]['data']
    nT = T.size

    # interpolate
    angle = braggT[None, :] + angle_rel[:, None]
    angle = np.linspace(np.min(angle), np.max(angle), 2*na + 1)
    pr = np.full((2*na+1, nT), np.nan)
    for ii in range(nT):
        pr[:, ii] = scpinterp.interp1d(
            braggT[ii] + angle_rel,
            power_ratio,
            kind='linear',
            fill_value=0,
            bounds_error=False,
        )(angle)

    coll2 = coll.__class__()
    coll2.add_ref(key='nT', size=nT)
    coll2.add_ref(key='na', size=2*na + 1)
    coll2.add_data(key='T', data=T, ref='nT', units='C')
    coll2.add_data(
        key='incidence angle absolute',
        data=angle,
        ref='na',
        units='rad',
    )
    coll2.add_data(
        key='power_ratio',
        data=pr,
        ref=('na', 'nT'),
        units='',
    )

    return coll2.plot_as_array(
        key='power_ratio',
        keyX='incidence angle absolute',
        keyY='T',
        dax=dax,
        aspect='auto',
    )



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import warnings
import itertools as itt

# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu.spectro as tfs

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.spectro'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


# def my_setup_function():
#    print ("my_setup_function")

# def my_teardown_function():
#    print ("my_teardown_function")

# @with_setup_method(my_setup_function, my_teardown_function)
# def test_numbers_3_4():
#    print 'test_numbers_3_4  <============================ actual test code'
#    assert multiply(3,4) == 12

# @with_setup_method(my_setup_function, my_teardown_function)
# def test_strings_a_3():
#    print 'test_strings_a_3  <============================ actual test code'
#    assert multiply('a',3) == 'aaa'


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_LocalExtrema(object):

    @classmethod
    def setup_class(cls):

        pfe = os.path.join(_here, 'test_data', 'UV_spectra_sh55506.npz')
        out = dict(np.load(pfe, allow_pickle=True))
        cls.lamb = out['lamb']
        cls.data = out['data']

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_get_local_extrema(self):

        lmethod = ['find_peaks', 'bspline']
        lreturnas = [bool, float]
        lreturn_min = [True, False]
        lreturn_prom = [True, False]

        for comb in itt.product(lmethod, lreturnas, lreturn_min, lreturn_prom):
            out = tfs.get_localextrema_1d(
                data=self.data[:10, :], lamb=self.lamb,
                width=None, weights=None,
                method=comb[0], returnas=comb[1],
                return_minima=comb[2],
                return_prominence=comb[3],
                return_width=None,
            )



from ._utils import convert_momentum_velocity_energy
from ._distribution import get_critical_dreicer_electric_fields
from ._distribution import get_normalized_momentum_distribution
from ._distribution import get_growth_source_terms
from . import emission



# -*- coding: utf-8 -*-


import warnings


import numpy as np
import bsplines2d as bs2
from contourpy import contour_generator
from matplotlib.path import Path
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt
import datastock as ds
import Polygon as plg


# ###############################################
# ###############################################
#           margin polygons
# ###############################################


def _get_poly_margin(
    # polygon
    p0=None,
    p1=None,
    # margin
    margin=None,
):

    # ----------
    # check

    margin = float(ds._generic_check._check_var(
        margin, 'margin',
        types=(float, int),
        default=0.3,
        sign='>0'
    ))

    # ---------------------------
    # add extra margin to pcross

    # get centroid
    cent = plg.Polygon(np.array([p0, p1]).T).center()

    # add margin
    return (
        cent[0] + (1. + margin) * (p0 - cent[0]),
        cent[1] + (1. + margin) * (p1 - cent[1]),
    )


# ###############################################
# ###############################################
#           overall polygons
# ###############################################


def _get_overall_polygons(
    coll=None,
    doptics=None,
    key_cam=None,
    poly=None,
    convexHull=None,
):

    # get temporary vos
    kp0, kp1 = doptics[key_cam]['dvos'][poly]
    shape = coll.ddata[kp0]['data'].shape
    p0 = coll.ddata[kp0]['data'].reshape((shape[0], -1))
    p1 = coll.ddata[kp1]['data'].reshape((shape[0], -1))

    # pix indices
    iok = np.isfinite(p0)

    # -----------------------
    # envelop pcross and phor

    if convexHull is True:
        # replace by convex hull
        pts = np.array([p0[iok], p1[iok]]).T
        return pts[ConvexHull(pts).vertices, :].T

    else:

        ipn = (np.all(iok, axis=0)).nonzero()[0]
        pp = plg.Polygon(np.array([p0[:, ipn[0]], p1[:, ipn[0]]]).T)
        for ii in ipn[1:]:
            pp |= plg.Polygon(np.array([p0[:, ii], p1[:, ii]]).T)

        if len(pp) > 1:

            # replace by convex hull
            pts = np.concatenate(
                tuple([np.array(pp.contour(ii)) for ii in range(len(pp))]),
                axis=0,
            )
            poly = pts[ConvexHull(pts).vertices, :].T

            # plot for debugging
            fig = plt.figure()
            fig.suptitle("_get_overall_polygons()", size=12)
            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
            ax.set_title(f"camera '{key_cam}', vos poly '{poly}'")
            for ii in range(len(pp)):
                ax.plot(
                    np.array(pp.contour(ii))[:, 0],
                    np.array(pp.contour(ii))[:, 1],
                    '.-',
                    poly[0, :],
                    poly[1, :],
                    '.-k',
                )
            msg = "multiple contours"
            warnings.warn(msg)
            return poly

        else:
            return np.array(pp.contour(0)).T


# ###############################################
# ###############################################
#               get polygons
# ###############################################


def _get_polygons(
    x0=None,
    x1=None,
    bool_cross=None,
    res=None,
):
    """ Get simplified contour polygon

    First computes contours
    Then simplifies it using a mix of convexHull and concave picking edges
    """

    # ------------
    # get contour

    contgen = contour_generator(
        x=x0,
        y=x1,
        z=bool_cross,
        name='serial',
        corner_mask=None,
        line_type='Separate',
        fill_type=None,
        chunk_size=None,
        chunk_count=None,
        total_chunk_count=None,
        quad_as_tri=True,       # for sub-mesh precision
        # z_interp=<ZInterp.Linear: 1>,
        thread_count=0,
    )

    no_cont, cj = bs2._class02_contours._get_contours_lvls(
        contgen=contgen,
        level=0.5,
        largest=True,
    )

    assert no_cont is False

    # -------------
    # simplify poly

    return _simplify_polygon(cj[:, 0], cj[:, 1], res=res)


def _simplify_polygon(c0, c1, res=None):

    # -----------
    # convex hull

    npts = c0.size

    # get hull
    convh = ConvexHull(np.array([c0, c1]).T)
    indh = convh.vertices
    ch0 = c0[indh]
    ch1 = c1[indh]
    nh = indh.size

    sign = np.median(np.diff(indh))

    # segments norms
    seg0 = np.r_[ch0[1:] - ch0[:-1], ch0[0] - ch0[-1]]
    seg1 = np.r_[ch1[1:] - ch1[:-1], ch1[0] - ch1[-1]]
    norms = np.sqrt(seg0**2 + seg1**2)

    # keep egdes that match res
    lind = []
    for ii, ih in enumerate(indh):

        # ind of points in between
        i1 = indh[(ii + 1) % nh]
        if sign > 0:
            if i1 > ih:
                ind = np.arange(ih, i1 + 1)
            else:
                ind = np.r_[np.arange(ih, npts), np.arange(0, i1 + 1)]
        else:
            if i1 < ih:
                ind = np.arange(ih, i1 - 1, -1)
            else:
                ind = np.r_[np.arange(ih, -1, -1), np.arange(npts - 1, i1 - 1, -1)]

        # trivial
        if ind.size == 2:
            lind.append((ih, i1))
            continue

        # get distances
        x0 = c0[ind]
        x1 = c1[ind]

        # segment unit vect
        vect0 = x0 - ch0[ii]
        vect1 = x1 - ch1[ii]

        # perpendicular distance
        cross = (vect0*seg1[ii] - vect1*seg0[ii]) / norms[ii]

        # criterion
        if np.all(np.abs(cross) <= 0.8*res):
            lind.append((ih, i1))
        else:
            lind += _simplify_concave(
                x0=x0,
                x1=x1,
                ind=ind,
                cross=cross,
                res=res,
            )

    # ------------------------------------
    # point by point on remaining segments

    iok = np.unique(np.concatenate(tuple(lind)))

    return c0[iok], c1[iok]


def _simplify_concave(
    x0=None,
    x1=None,
    ind=None,
    cross=None,
    res=None,
):

    # ------------
    # safety check

    imax = np.argmax(np.abs(cross))
    sign0 = np.sign(cross[imax])
    # sign0 = np.mean(sign)
    iok0 = cross * sign0 >= -1e-12
    if not np.all(iok0):
        fig = plt.figure()
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.plot(
            x0, x1, '.-k',
        )
        msg = (
            "Non-conform cross * sign0 (>= -1e-12):\n"
            f"\t- x0 = {x0}\n"
            f"\t- x1 = {x1}\n"
            f"\t- ind = {ind}\n"
            f"\t- iok0 = {iok0}\n"
            # f"\t- sign = {sign}\n"
            f"\t- sign0 = {sign0}\n"
            f"\t- cross * sign0 = {cross * sign0}\n"
        )
        raise Exception(msg)

    # ------------
    # loop

    i0 = 0
    i1 = 1
    iok = 1
    lind_loc, lind = [], []
    while iok <= ind.size - 1:

        # reference normalized vector
        vref0, vref1 = x0[i1] - x0[i0], x1[i1] - x1[i0]
        normref = np.sqrt(vref0**2 + vref1**2)
        vref0, vref1 = vref0 / normref, vref1 / normref

        # intermediate vectors
        indi = np.arange(i0 + 1, i1)
        v0 = x0[indi] - x0[i0]
        v1 = x1[indi] - x1[i0]

        # sign and distance (from cross product)
        cross = v0 * vref1 - v1 * vref0
        dist = np.abs(cross)

        # conditions
        c0 = np.all(dist <= 0.8*res)
        c1 = np.all(cross * sign0 >= -1e-12)
        c2 = i1 == ind.size - 1

        append = False
        # cases
        if c0 and c1 and (not c2):
            iok = int(i1)
            i1 += 1
        elif c0 and c1 and c2:
            iok = int(i1)
            append = True
        elif c0 and (not c1) and (not c2):
            i1 += 1
        elif c0 and (not c1) and c2:
            append = True
        elif not c0:
            append = True

        # append
        if append is True:
            lind_loc.append((i0, iok))
            lind.append((ind[i0], ind[iok]))
            i0 = iok
            i1 = i0 + 1
            iok = int(i1)

        if i1 > ind.size - 1:
            break

    return lind


# ################################################################
# ################################################################
#               Get dphi from R and phor
# ################################################################


def _get_dphi_from_R_phor(
    R=None,
    phor0=None,
    phor1=None,
    phimin=None,
    phimax=None,
    res=None,
    out=None
):

    # ------------
    # check inputs

    # out
    out = ds._generic_check._check_var(
        out, 'out',
        types=bool,
        default=True,
    )
    sign = 1. if out is True else -1.

    # R
    R = np.unique(np.atleast_1d(R).ravel())

    # path
    path = Path(np.array([phor0, phor1]).T)

    # --------------
    # sample phi

    dphi = np.full((2, R.size), np.nan)
    for ir, rr in enumerate(R):

        nphi = np.ceil(rr*np.abs(phimax - phimin) / (0.05*res)).astype(int) + 1
        phi = np.linspace(phimin, phimax, nphi)

        ind = path.contains_points(
            np.array([rr*np.cos(phi), rr*np.sin(phi)]).T
        )

        if np.any(ind):
            dphi[0, ir] = np.min(phi[ind]) - sign*(phi[1] - phi[0])
            dphi[1, ir] = np.max(phi[ind]) + sign*(phi[1] - phi[0])

    return dphi